import numpy as np
import tensorflow as tf

from .modelparams import ModelParams
from .custom_ops import n_dimensional_weightmul
from .custom_ops import gaussian_log_likelihood


class VAEParams(ModelParams):
    def __init__(self):
        self.INPUT_SHAPE = [28, 28, 1, 1]
        self.CONVOLUTION_LAYERS = [
            {
                "type": "conv3d",
                "filter": [5, 5, 1, 1, 64],
                "downsampling": {"type": "max_pool3d", "k": 2},
            },
            {
                "type": "conv3d",
                "filter": [3, 3, 1, 64, 128],
                "downsampling": {"type": "max_pool3d", "k": 2},
            },
            {
                "type": "conv3d",
                "filter": [3, 3, 1, 128, 256],
                "downsampling": {"type": "max_pool3d", "k": 2},
            },
        ]
        self.HIDDEN_LAYERS = [{"shape": [1000]}, {"shape": [600]}, {"shape": [400]}]
        self.LATENT_SHAPE = [15]
        self.COERCED_LATENT_DIMS = 5
        self.LEARNING_RATE = 0.00001
        self.DROPOUT = 0.8  # Keep-prob
        self.FLOAT_TYPE = tf.float32
        self.DISABLE_SUMMARY = False
        self.ADVERSARIAL = False
        self.MUTUAL_INFO = False
        self.INFO_REG_COEFF = 0.5


TINY = 1e-8


class Autoencoder(object):
    def __init__(self, model_params):
        self.MP = model_params

        tf.reset_default_graph()
        preset_batch_size = None
        self.variables = []
        self.G_variables = []
        self.Q_only_variables = []
        self.Q_and_D_variables = []
        self.D_only_variables = []
        self.zero = tf.constant(0)
        # Graph input
        with tf.name_scope("Placeholders") as scope:
            self.input_placeholder = tf.placeholder(
                self.MP.FLOAT_TYPE,
                shape=[preset_batch_size] + self.MP.INPUT_SHAPE,
                name="input",
            )
            if self.MP.DROPOUT is not None:
                default_dropout = tf.constant(1, dtype=self.MP.FLOAT_TYPE)
                self.dropout_placeholder = tf.placeholder_with_default(
                    default_dropout, (), name="dropout_prob"
                )
            self.stop_gradient_placeholder = tf.placeholder_with_default(
                tf.constant(False, dtype=tf.bool), (), name="stop_gradient_at_z"
            )
        # Encoder
        previous_layer = self.input_placeholder
        previous_layer_shape = (
            self.MP.INPUT_SHAPE
        )  # Excludes batch dim (which should be at pos 0)
        # Convolutional Layers
        self.conv_layers_input_shapes = []
        for i, LAYER in enumerate(self.MP.CONVOLUTION_LAYERS):
            self.conv_layers_input_shapes.append(previous_layer_shape)
            previous_layer, previous_layer_shape = self.build_conv_layer(
                LAYER,
                previous_layer,
                previous_layer_shape,
                "ConvLayer" + str(i),
                "ConvLayer" + str(i) + "Weights",
                "_conv_" + str(i),
                put_variables_in_list=self.Q_and_D_variables,
            )
        # Flatten output
        self.shape_before_flattening = previous_layer_shape
        previous_layer_shape = [np.prod(previous_layer_shape)]
        previous_layer = tf.reshape(
            previous_layer, shape=[-1] + previous_layer_shape, name="flatten"
        )
        # Fully connected Layers
        for i, LAYER in enumerate(self.MP.HIDDEN_LAYERS):
            previous_layer, previous_layer_shape = self.build_FC_layer(
                LAYER,
                previous_layer,
                previous_layer_shape,
                "EncoderLayer" + str(i),
                "EncoderLayer" + str(i) + "Weights",
                "_encoder_" + str(i),
                put_variables_in_list=self.Q_only_variables,
            )
        # Latent space
        self.z_mean, _ = self.build_FC_layer(
            {"shape": self.MP.LATENT_SHAPE},
            previous_layer,
            previous_layer_shape,
            "ZMeanLayer",
            "LatentLayerWeights",
            "_z_mean",
            activation=lambda x: x,
            put_variables_in_list=self.Q_only_variables,
        )
        self.z_log_sigma_squared, _ = self.build_FC_layer(
            {"shape": self.MP.LATENT_SHAPE},
            previous_layer,
            previous_layer_shape,
            "ZLogSigmaSquaredLayer",
            "LatentLayerWeights",
            "_z_log_sig2",
            activation=lambda x: x,
            put_variables_in_list=self.Q_only_variables,
        )
        # Sample Z values from Latent-space Estimate
        with tf.name_scope("SampleZValues") as scope:
            # sample = mean + sigma*epsilon
            epsilon = tf.random_normal(
                tf.shape(self.z_mean),
                0,
                1,
                dtype=self.MP.FLOAT_TYPE,
                name="randomnormal",
            )
            self.z_sample = tf.add(
                self.z_mean,
                tf.multiply(tf.sqrt(tf.exp(self.z_log_sigma_squared)), epsilon),
                name="z_sample",
            )
        # Stop gradients if desired
        previous_layer = tf.cond(
            self.stop_gradient_placeholder,
            lambda: tf.stop_gradient(self.z_sample),
            lambda: self.z_sample,
        )
        # Decoder (a.k.a Generator)
        # Fully connected layers
        previous_layer = previous_layer
        previous_layer_shape = self.MP.LATENT_SHAPE
        for i, LAYER in enumerate(self.MP.HIDDEN_LAYERS[::-1]):
            previous_layer, previous_layer_shape = self.build_FC_layer(
                LAYER,
                previous_layer,
                previous_layer_shape,
                "DecoderLayer" + str(i),
                "DecoderLayer" + str(i) + "Weights",
                "_decoder_" + str(i),
                put_variables_in_list=self.G_variables,
            )
        # Post fully-connected layer
        previous_layer, previous_layer_shape = self.build_FC_layer(
            {"shape": [np.prod(self.shape_before_flattening)]},
            previous_layer,
            previous_layer_shape,
            "Reconstruction",
            "ReconstructionLayerWeights",
            "_reconstruction",
            activation=tf.nn.sigmoid,
            put_variables_in_list=self.G_variables,
        )
        # Unflatten output
        previous_layer = tf.reshape(
            previous_layer, shape=[-1] + self.shape_before_flattening, name="unflatten"
        )
        previous_layer_shape = self.shape_before_flattening
        # Deconvolutional Layers
        deconv_layers_output_shapes = self.conv_layers_input_shapes[::-1]
        dynamic_batch_size = tf.shape(previous_layer)[0]
        for i, (LAYER, target_output_shape) in enumerate(
            zip(self.MP.CONVOLUTION_LAYERS[::-1], deconv_layers_output_shapes)
        ):
            previous_layer, previous_layer_shape = self.build_deconv_layer(
                LAYER,
                previous_layer,
                previous_layer_shape,
                target_output_shape,
                dynamic_batch_size,
                "DeConvLayer" + str(i),
                "DeConvLayer" + str(i) + "Weights",
                "_deconv_" + str(i),
                put_variables_in_list=self.G_variables,
            )
        # Output (as probability of output being 1)
        self.output = tf.minimum(previous_layer, 1)
        # Loss
        with tf.name_scope("Losses") as scope:
            with tf.name_scope("ReconstructionLoss") as sub_scope:
                # Cross entropy loss of output probabilities vs. input certainties.
                self.reconstruction_loss = -tf.reduce_mean(
                    self.input_placeholder * tf.log(TINY + self.output, name="log1")
                    + (1 - self.input_placeholder)
                    * tf.log(TINY + (1 - self.output), name="log2")
                )
            with tf.name_scope("LatentLoss") as sub_scope:
                # Kullback Leibler divergence between latent normal distribution and ideal.
                self.latent_loss = (
                    -0.5
                    * 0.001
                    * tf.reduce_mean(
                        1
                        + self.z_log_sigma_squared
                        - tf.square(self.z_mean)
                        - tf.exp(self.z_log_sigma_squared)
                    )
                )
            # Average sum of costs over batch.
            self.cost = self.reconstruction_loss + self.latent_loss
            if not self.MP.DISABLE_SUMMARY:
                tf.summary.scalar("self.reconstruction_loss", self.reconstruction_loss)
                tf.summary.scalar("self.latent_loss", self.latent_loss)
                tf.summary.scalar("autoencoder_loss", self.cost)
        # Extra graph
        if self.MP.ADVERSARIAL:
            self.build_adversarial_graph()
        if self.MP.MUTUAL_INFO:
            self.build_mutual_info_graph()
        # Optimizers (ADAM)
        with tf.name_scope("Optimizer") as scope:
            self.optimizer = tf.train.AdamOptimizer(
                learning_rate=self.MP.LEARNING_RATE
            ).minimize(self.cost)
            if self.MP.MUTUAL_INFO:
                self.optimizer_with_MI = tf.train.AdamOptimizer(
                    learning_rate=self.MP.LEARNING_RATE
                ).minimize(self.cost_with_MI)
        if self.MP.ADVERSARIAL:
            with tf.name_scope("Adversarial_Optimizers") as scope:
                with tf.name_scope("Generator_Optimizer") as sub_scope:
                    self.generator_optimizer = tf.train.AdamOptimizer(
                        learning_rate=self.MP.LEARNING_RATE
                    ).minimize(
                        self.generator_loss,
                        var_list=self.G_variables
                        + self.Q_only_variables
                        + self.Q_and_D_variables,
                    )
                with tf.name_scope("Discriminator_Optimizer") as sub_scope:
                    self.discriminator_optimizer = tf.train.AdamOptimizer(
                        learning_rate=self.MP.LEARNING_RATE
                    ).minimize(
                        self.discriminator_loss,
                        var_list=self.D_only_variables + self.Q_and_D_variables,
                    )
        # Initialize session
        self.catch_nans = tf.add_check_numerics_ops()
        self.sess = tf.Session()
        self.merged = (
            tf.summary.merge_all() if not self.MP.DISABLE_SUMMARY else self.zero
        )
        tf.global_variables_initializer().run(session=self.sess)
        # Saver
        self.saver = tf.train.Saver(self.variables)
        tf.get_default_graph().finalize()

    def build_adversarial_graph(self):
        print("Building adversarial graph.")
        with tf.name_scope("Adversarial") as meta_scope:
            ## Apply Discriminator to real x
            previous_layer = self.input_placeholder
            previous_layer_shape = self.MP.INPUT_SHAPE
            # Convolutional Layers
            for i, LAYER in enumerate(self.MP.CONVOLUTION_LAYERS):
                previous_layer, previous_layer_shape = self.build_conv_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Discriminator_ConvLayer" + str(i),
                    "ConvLayer" + str(i) + "Weights",
                    "_conv_" + str(i),
                    reuse=True,
                )
            # Flatten output
            previous_layer_shape = [np.prod(previous_layer_shape)]
            previous_layer = tf.reshape(
                previous_layer, shape=[-1] + previous_layer_shape, name="flatten"
            )
            # Fully connected Layers
            for i, LAYER in enumerate(self.MP.HIDDEN_LAYERS):
                previous_layer, previous_layer_shape = self.build_FC_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Discriminator_EncoderLayer" + str(i),
                    "DiscriminatorFCLayer" + str(i) + "Weights",
                    "_discriminator_fc_" + str(i),
                    reuse=False,
                    put_variables_in_list=self.D_only_variables,
                )
            # Latent space
            self.discriminator_output_real, _ = self.build_FC_layer(
                {"shape": [1]},
                previous_layer,
                previous_layer_shape,
                "Discriminator_output",
                "DiscriminatorOutputWeights",
                reuse=False,
                activation=tf.nn.sigmoid,
                put_variables_in_list=self.D_only_variables,
            )
            ## Apply Discriminator to generator output
            previous_layer = self.output
            previous_layer_shape = self.MP.INPUT_SHAPE
            # Convolutional Layers
            for i, LAYER in enumerate(self.MP.CONVOLUTION_LAYERS):
                previous_layer, previous_layer_shape = self.build_conv_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Discriminator_ConvLayer" + str(i),
                    "ConvLayer" + str(i) + "Weights",
                    "_conv_" + str(i),
                    reuse=True,
                )
            # Flatten output
            previous_layer_shape = [np.prod(previous_layer_shape)]
            previous_layer = tf.reshape(
                previous_layer, shape=[-1] + previous_layer_shape, name="flatten"
            )
            # Fully connected Layers
            for i, LAYER in enumerate(self.MP.HIDDEN_LAYERS):
                previous_layer, previous_layer_shape = self.build_FC_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Discriminator_EncoderLayer" + str(i),
                    "DiscriminatorFCLayer" + str(i) + "Weights",
                    "_discriminator_fc_" + str(i),
                    reuse=True,
                )
            # Latent space
            self.discriminator_output_fake, _ = self.build_FC_layer(
                {"shape": [1]},
                previous_layer,
                previous_layer_shape,
                "Discriminator_output",
                "DiscriminatorOutputWeights",
                reuse=True,
                activation=tf.nn.sigmoid,
            )
            # Loss
            with tf.name_scope("Adversarial_Loss") as scope:
                self.discriminator_loss = tf.reduce_mean(
                    0.5 * -tf.log(self.discriminator_output_real + TINY)
                    + 0.5
                    * -tf.log(
                        tf.maximum((1.0 - self.discriminator_output_fake), 0.) + TINY,
                        name="logDfake",
                    )
                )
                self.generator_loss = tf.reduce_mean(
                    -tf.log(self.discriminator_output_fake + TINY)
                )
                self.discriminator_loss_no_MI = self.discriminator_loss * 1.
                self.generator_loss_no_MI = self.generator_loss * 1.
        with tf.name_scope("Losses") as meta_scope:
            if not self.MP.DISABLE_SUMMARY:
                tf.summary.scalar("generator_loss", self.generator_loss_no_MI)
                tf.summary.scalar("discriminator_loss", self.discriminator_loss_no_MI)

    def build_mutual_info_graph(self):
        print("Building mutual information graph.")
        with tf.name_scope("Mutual_Info") as meta_scope:
            # Encoder
            previous_layer = self.output
            previous_layer_shape = (
                self.MP.INPUT_SHAPE
            )  # Excludes batch dim (which should be at pos 0)
            # Convolutional Layers
            for i, LAYER in enumerate(self.MP.CONVOLUTION_LAYERS):
                previous_layer, previous_layer_shape = self.build_conv_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Q_ConvLayer" + str(i),
                    "ConvLayer" + str(i) + "Weights",
                    "_conv_" + str(i),
                    reuse=True,
                )
            # Flatten output
            previous_layer_shape = [np.prod(previous_layer_shape)]
            previous_layer = tf.reshape(
                previous_layer, shape=[-1] + previous_layer_shape, name="flatten"
            )
            # Fully connected Layers
            for i, LAYER in enumerate(self.MP.HIDDEN_LAYERS):
                previous_layer, previous_layer_shape = self.build_FC_layer(
                    LAYER,
                    previous_layer,
                    previous_layer_shape,
                    "Q_EncoderLayer" + str(i),
                    "EncoderLayer" + str(i) + "Weights",
                    "_encoder_" + str(i),
                    reuse=True,
                )
            # Latent space
            self.q_z_mean, _ = self.build_FC_layer(
                {"shape": self.MP.LATENT_SHAPE},
                previous_layer,
                previous_layer_shape,
                "Q_ZMeanLayer",
                "LatentLayerWeights",
                "_z_mean",
                activation=lambda x: x,
                reuse=True,
            )
            self.q_z_log_sigma_squared, _ = self.build_FC_layer(
                {"shape": self.MP.LATENT_SHAPE},
                previous_layer,
                previous_layer_shape,
                "Q_ZLogSigmaSquaredLayer",
                "LatentLayerWeights",
                "_z_log_sig2",
                activation=lambda x: x,
                reuse=True,
            )
            self.q_z_mean = self.q_z_mean[:, : self.MP.COERCED_LATENT_DIMS]
            self.q_z_log_sigma_squared = self.q_z_log_sigma_squared[
                :, : self.MP.COERCED_LATENT_DIMS
            ]
            # Loss
            with tf.name_scope("MutualInfo_Loss") as scope:
                c_sample = self.z_sample[:, : self.MP.COERCED_LATENT_DIMS]
                c_mean_prior = self.z_mean[:, : self.MP.COERCED_LATENT_DIMS]
                c_log_sigma_squared_prior = self.z_log_sigma_squared[
                    :, : self.MP.COERCED_LATENT_DIMS
                ]
                log_li_q_c_given_x = gaussian_log_likelihood(
                    c_sample, self.q_z_mean, self.q_z_log_sigma_squared
                )
                log_li_q_c = gaussian_log_likelihood(
                    c_sample, c_mean_prior, c_log_sigma_squared_prior
                )
                self.mutual_information_est = tf.reduce_mean(
                    -log_li_q_c
                ) - tf.reduce_mean(-log_li_q_c_given_x)
                if self.MP.ADVERSARIAL:
                    self.discriminator_loss -= (
                        self.MP.INFO_REG_COEFF * self.mutual_information_est
                    )
                    self.generator_loss -= (
                        self.MP.INFO_REG_COEFF * self.mutual_information_est
                    )
                self.cost_with_MI = (
                    self.cost
                    - 0.01 * self.MP.INFO_REG_COEFF * self.mutual_information_est
                )
        with tf.name_scope("Losses") as meta_scope:
            if not self.MP.DISABLE_SUMMARY:
                if self.MP.ADVERSARIAL:
                    tf.summary.scalar("generator_loss_with_MI", self.generator_loss)
                    tf.summary.scalar(
                        "discriminator_loss_with_MI", self.discriminator_loss
                    )
                tf.summary.scalar("autoencoder_loss_with_MI", self.cost_with_MI)
                tf.summary.scalar("MI", self.mutual_information_est)

    def variable_summaries(self, var):
        """Attach a lot of summaries to a Tensor (for TensorBoard visualization)."""
        if not self.MP.DISABLE_SUMMARY:
            with tf.name_scope("summaries"):
                mean = tf.reduce_mean(var)
                tf.summary.scalar("mean", mean)
                with tf.name_scope("stddev"):
                    stddev = tf.sqrt(tf.reduce_mean(tf.square(var - mean)))
                tf.summary.scalar("stddev", stddev)
                tf.summary.scalar("max", tf.reduce_max(var))
                tf.summary.scalar("min", tf.reduce_min(var))
                tf.summary.histogram("histogram", var)

    def build_conv_layer(
        self,
        LAYER,
        previous_layer,
        previous_layer_shape,
        scope_name,
        varscope_name,
        var_suffix="",
        reuse=False,
        put_variables_in_list=None,
    ):
        with tf.name_scope(scope_name) as scope:
            filter_shape = LAYER["filter"]
            stride = LAYER["stride"] if "stride" in LAYER else 1
            strides = [1, stride, stride, stride, 1]
            padding = LAYER["padding"] if "padding" in LAYER else "SAME"
            with tf.variable_scope(varscope_name, reuse=reuse) as varscope:
                weights = tf.get_variable(
                    "weights" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=filter_shape,
                    initializer=tf.contrib.layers.xavier_initializer(),
                )
                biases = tf.get_variable(
                    "biases_conv" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=[filter_shape[-1]],
                    initializer=tf.constant_initializer(0),
                )
            if not reuse:
                self.variables.append(weights)
                self.variables.append(biases)
                if put_variables_in_list is not None:
                    put_variables_in_list.append(weights)
                    put_variables_in_list.append(biases)
            layer_output = tf.nn.conv3d(previous_layer, weights, strides, padding)
            layer_output = tf.nn.bias_add(layer_output, biases)
            layer_shape = previous_layer_shape[:]
            for i, (prev_dim, filt_dim) in enumerate(
                zip(previous_layer_shape[:3], filter_shape[:3])
            ):
                pad = np.floor(filt_dim / 2) if padding == "SAME" else 0
                layer_shape[i] = int(((prev_dim + 2 * pad - filt_dim) / stride) + 1)
            layer_shape[-1] = filter_shape[-1]
            # Downsampling
            if "downsampling" in LAYER:
                DOWNSAMPLING = LAYER["downsampling"]
                if DOWNSAMPLING["type"] != "max_pool3d":
                    raise NotImplementedError
                if self.MP.FLOAT_TYPE != tf.float32:
                    raise TypeError("max_pool3d only supports float32")
                k = DOWNSAMPLING["k"] if "k" in DOWNSAMPLING else 2
                ksize = [1, k, k, k, 1]
                strides = [1, k, k, k, 1]
                padding = (
                    DOWNSAMPLING["padding"] if "padding" in DOWNSAMPLING else "VALID"
                )
                layer_output = tf.nn.max_pool3d(layer_output, ksize, strides, padding)
                pad = np.floor(k / 2) if padding == "SAME" else 0
                layer_shape = [
                    int(((dim + 2 * pad - k) / k) + 1) for dim in layer_shape[:3]
                ] + layer_shape[3:]
        return layer_output, layer_shape

    def build_FC_layer(
        self,
        LAYER,
        previous_layer,
        previous_layer_shape,
        scope_name,
        varscope_name,
        var_suffix="",
        reuse=False,
        activation=tf.nn.softplus,
        put_variables_in_list=None,
    ):
        with tf.name_scope(scope_name) as scope:
            layer_shape = LAYER["shape"]
            with tf.variable_scope(varscope_name, reuse=reuse) as varscope:
                weights = tf.get_variable(
                    "weights" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=previous_layer_shape + layer_shape,
                    initializer=tf.contrib.layers.xavier_initializer(),
                )
                biases = tf.get_variable(
                    "biases" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=layer_shape,
                    initializer=tf.constant_initializer(0),
                )
            if not reuse:
                self.variables.append(weights)
                self.variables.append(biases)
                if put_variables_in_list is not None:
                    put_variables_in_list.append(weights)
                    put_variables_in_list.append(biases)
                self.variable_summaries(weights)
            layer_output = tf.add(
                n_dimensional_weightmul(
                    previous_layer, weights, previous_layer_shape, layer_shape
                ),
                biases,
            )
            layer_output = activation(layer_output)
            if self.MP.DROPOUT is not None:
                layer_output = tf.nn.dropout(layer_output, self.dropout_placeholder)
        return layer_output, layer_shape

    def build_deconv_layer(
        self,
        LAYER,
        previous_layer,
        previous_layer_shape,
        target_output_shape,
        dynamic_batch_size,
        scope_name,
        varscope_name,
        var_suffix="",
        reuse=False,
        put_variables_in_list=None,
    ):
        with tf.name_scope(scope_name) as scope:
            filter_shape = LAYER["filter"][:]
            stride = LAYER["downsampling"]["k"] if "downsampling" in LAYER else 1
            strides = [1, stride, stride, stride, 1]
            padding = LAYER["padding"] if "padding" in LAYER else "SAME"
            with tf.variable_scope(varscope_name, reuse=reuse) as varscope:
                weights = tf.get_variable(
                    "weights" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=filter_shape,
                    initializer=tf.contrib.layers.xavier_initializer(),
                )
                biases = tf.get_variable(
                    "biases" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=[filter_shape[-2]],
                    initializer=tf.constant_initializer(0),
                )
            if not reuse:
                self.variables.append(weights)
                self.variables.append(biases)
                if put_variables_in_list is not None:
                    put_variables_in_list.append(weights)
                    put_variables_in_list.append(biases)
            output_shape = [dynamic_batch_size] + target_output_shape
            layer_output = tf.nn.conv3d_transpose(
                previous_layer,
                weights,
                output_shape=output_shape,
                strides=strides,
                padding=padding,
            )
            layer_output = tf.nn.bias_add(layer_output, biases)
            layer_output = tf.nn.relu(layer_output)
            layer_shape = output_shape[1:]
        return layer_output, layer_shape

    ## Example functions for different ways to call the autoencoder graph.
    def encode(self, batch_input):
        return self.sess.run(
            (self.z_mean, self.z_log_sigma_squared),
            feed_dict={self.input_placeholder: batch_input},
        )

    def decode(self, batch_z):
        return self.sess.run(self.output, feed_dict={self.z_sample: batch_z})

    def encode_decode(self, batch_input):
        return self.sess.run(
            self.output, feed_dict={self.input_placeholder: batch_input}
        )

    def train_on_single_batch(
        self,
        batch_input,
        train_target=None,
        cost_only=False,
        dropout=None,
        summary_writer=None,
    ):
        # feed placeholders
        dict_ = {self.input_placeholder: batch_input}
        if self.MP.DROPOUT is not None:
            dict_[self.dropout_placeholder] = (
                self.MP.DROPOUT if dropout is None else dropout
            )
        else:
            if dropout is not None:
                raise ValueError(
                    "This model does not implement dropout yet a value was specified"
                )
        # Graph nodes to target
        cost = [self.cost]
        if self.MP.ADVERSARIAL:
            cost = cost + [self.generator_loss_no_MI, self.discriminator_loss_no_MI]
        if self.MP.MUTUAL_INFO:
            cost = cost + [self.mutual_information_est]
        opt = train_target if train_target is not None else self.optimizer
        if self.MP.ADVERSARIAL:
            if opt is self.discriminator_optimizer or opt is self.generator_optimizer:
                dict_[self.stop_gradient_placeholder] = True
        if self.MP.MUTUAL_INFO:
            if opt is self.optimizer_with_MI:
                dict_[self.stop_gradient_placeholder] = True
        # compute
        cost, _, _, summary = self.sess.run(
            (cost, opt, self.catch_nans, self.merged), feed_dict=dict_
        )
        if summary_writer is not None:
            summary_writer.add_summary(summary)
        return np.array(cost)

    def cost_on_single_batch(self, batch_input, summary_writer=None):
        return self.train_on_single_batch(
            batch_input,
            train_target=self.zero,
            dropout=1.0,
            summary_writer=summary_writer,
        )

    def batch_encode(self, batch_input, batch_size=200, verbose=True):
        return batch_generic_func(self.encode, batch_input, batch_size, verbose)

    def batch_decode(self, batch_input, batch_size=200, verbose=True):
        return batch_generic_func(self.decode, batch_input, batch_size, verbose)

    def batch_encode_decode(self, batch_input, batch_size=100, verbose=True):
        return batch_generic_func(self.encode_decode, batch_input, batch_size, verbose)


def concat(a, b):
    if isinstance(a, list):
        return a + b
    elif isinstance(b, np.ndarray):
        return np.concatenate((a, b), axis=0)
    else:
        raise TypeError("Inputs are of unsupported type")


def batch_add(A, B):
    if isinstance(A, tuple):
        return tuple([concat(a, b) for a, b in zip(A, B)])
    else:
        return concat(A, B)


def batch_generic_func(function, batch_input, batch_size=100, verbose=False):
    a = 0
    b = batch_size
    while a < len(batch_input):
        batch_output = function(batch_input[a:b])
        try:
            output = batch_add(output, batch_output)
        except (NameError):
            output = batch_output
        a += batch_size
        b += batch_size
        if verbose:
            print(
                "Example " + str(min(a, len(batch_input))) + "/" + str(len(batch_input))
            )
    return output



import matplotlib

def movefigure(fig, pos):
    """ Moves pyplot figure window to new position where top left corner is at pixel coordinates (px, py)"""
    px, py = pos
    backend = matplotlib.get_backend()
    if backend == 'WXAgg':
        fig.canvas.manager.window.SetPosition((px, py))
    elif backend in ['GTK', 'GTKAgg', 'GTKCairo', 'GTK3Agg', 'GTK3Cairo', 'Qt4Agg', 'Qt4Cairo', 'Qt5Agg', 'Qt5Cairo']:
        fig.canvas.manager.window.move(px, py)
    elif backend == 'TkAgg':
        fig.canvas.manager.window.wm_geometry("+{}+{}".format(px, py))
    else:
        raise NotImplementedError



### Tools for quickly getting ros messages

import functools

import rospy
import tf
from tf2_ros import TransformException
from sensor_msgs.msg import *
from std_msgs.msg import *
import numpy as np
from matplotlib import pyplot as plt
import sys

class QuickSubscriber(object):
    """ Quickly extracts messages """
    def __init__(self, topic_name=None, msg_class=Image, n_messages=1, verbose=False,
            tf_parent_frame=None, tf_timeout=1., tf_skip_if_no_tf=True):
        if topic_name is None:
            print("Usage: q = QuickSubscriber(topic_name, msg_class ,n_messages(default=1), verbose(default=False))")
            raise Exception()
        rospy.init_node("quick_subscriber_{}".format(topic_name.replace('/','_')), anonymous=True)
        rospy.Subscriber(topic_name, msg_class, self.callback)
        self.messages = []
        self.n_messages = n_messages
        self.topic_name = topic_name
        self.VERBOSE = verbose
        self.TF_PARENT_FRAME = tf_parent_frame
        self.tf_timeout = rospy.Duration(tf_timeout)
        self.tf_skip_if_tf_not_found = tf_skip_if_no_tf
        if self.TF_PARENT_FRAME is not None:
            self.tfs = []
            self.tf_listener = tf.TransformListener()
        print("Waiting for {} messages on topic {}".format(self.n_messages, self.topic_name))
        rospy.spin()

    def callback(self, msg):
        verbose_msg_info = ": {}".format(msg) if self.VERBOSE else ""
        rospy.loginfo("Captured message {}/{}{}".format(len(self.messages)+1,
                                                          self.n_messages,
                                                          verbose_msg_info));
        if self.TF_PARENT_FRAME is not None:
            try:
                time = rospy.Time(msg.header.stamp.secs, msg.header.stamp.nsecs)
                tf_info = [self.TF_PARENT_FRAME, msg.header.frame_id, time]
                self.tf_listener.waitForTransform(*(tf_info + [self.tf_timeout]))
                tf_ = self.tf_listener.lookupTransform(*tf_info)
            except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException,
                    TransformException) as e:
                print("tf for time {} not found: {}".format(time, e))
                if self.tf_skip_if_tf_not_found:
                    print("Skipping message.")
                    return
                else:
                    tf_ = None
            self.tfs.append(tf_)
        self.messages.append(msg)
        if len(self.messages) >= self.n_messages:
            print("Collected all messages. Messages stored in QuickSubscriber.messages")
            if self.TF_PARENT_FRAME is not None:
                print("tfs stored in QuickSubscriber.tfs")
            rospy.signal_shutdown("Mission accomplished.")

class QuickMultiSubscriber(object):
    """ Quickly extracts messages from multiple topics"""
    def __init__(self, subscribers_list=None, n_messages=1, verbose=False):
        if subscribers_list is None:
            print("Usage: q = QuickSubscriber(subscribers_list, n_messages(default=1), verbose(default=False))")
            print("     subscribers_list is a list of dicts, for example:")
            print("     [{'topic': 'my_topic', 'msg_class': sensor_msgs.msg.Image}]")
            raise Exception()
        self.topic_names = [item['topic'] for item in subscribers_list]
        self.messages = [[] for item in subscribers_list]
        self.msg_classes = [item['msg_class'] for item in subscribers_list]
        self.n_messages = n_messages
        self.VERBOSE = verbose
        rospy.init_node("quick_multi_subscriber_{}".format(self.topic_names[0].replace('/','_')), anonymous=True)
        for topic_name, msg_class, messages in zip(self.topic_names, self.msg_classes, self.messages):
            rospy.Subscriber(topic_name, msg_class, functools.partial(self.callback,
                                                                      messages,
                                                                      topic_name))
            print("Waiting for {} messages on topic {}".format(self.n_messages, topic_name))
        rospy.spin()

    def callback(self, messages, topic_name, msg):
        verbose_msg_info = ": {}".format(msg) if self.VERBOSE else ""
        rospy.loginfo("Captured message {}/{}{} for topic {}".format(len(messages),
                                                          self.n_messages,
                                                          verbose_msg_info,
                                                          topic_name))
        messages.append(msg)
        if len(messages) > self.n_messages:
            print("Collected all messages. Messages stored in QuickMultiSubscriber.messages")
            rospy.signal_shutdown("Mission accomplished.")

def subscribers_list_example():
    return [{'topic': "topic_a", 'msg_class': sensor_msgs.msg.Image},
            {'topic': "topic_b", 'msg_class': sensor_msgs.msg.Image}]





import rospy
import traceback

def shutdown_if_fail(function):
    def wrapper_for_routine(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except: # noqa
            traceback.print_exc()
            rospy.signal_shutdown("Exception caught in routine thread.")
            raise
    return wrapper_for_routine




import matplotlib.pyplot as plt
import numpy as np

# copied from map2d.py in danieldugas/pymap2d
def gridshow(*args, **kwargs):
    """ utility function for showing 2d grids in matplotlib,
    wrapper around pyplot.imshow

    use 'extent' = [-1, 1, -1, 1] to change the axis values """
    from matplotlib import pyplot as plt
    if not 'origin' in kwargs:
        kwargs['origin'] = 'lower'
    if not 'cmap' in kwargs:
        kwargs['cmap'] = plt.cm.Greys
    return plt.imshow(*(arg.T if i == 0 else arg for i, arg in enumerate(args)), **kwargs)

def plotstep(*args, **kwargs):
    if 'where' not in kwargs:
        kwargs['where'] = 'mid'
    # add x dim if missing
    x = None
    y = None
    newargs = []
    for i, arg in enumerate(args):
        if type(arg) in [np.ndarray, list]:
            if i == 0:
                x = arg
                continue
            if i == 1:
                y = arg
                continue
        newargs.append(arg)
    if y is None and x is not None:
        y = x
        x = np.arange(len(y))
        args = [x, y] + newargs
    # plot
    plt.step(*args, **kwargs)




import numpy as np

_range = range


def histogram_along_axis(a, axis=None, bins=10, range=None, weights=None):
    """ Get a histogram of values along specific axis, all sharing the same bins
    last axis in output should have size bins, and contains histogram values
    - Daniel Dugas

    See numpy.histogram

    Example
    -------
    >>> a = np.ones((2,2))
    >>> a
    array([[1., 1.],
           [1., 1.]])
    >>> histogram_along_axis(a, bins=4, axis=0)
    (array([[0, 0, 2, 0],
           [0, 0, 2, 0]]), array([0.5 , 0.75, 1.  , 1.25, 1.5 ]))
    """
    _, bin_edges = np.histogram(
        np.array([np.min(a), np.max(a)]), bins=bins, range=range, weights=weights
    )

    axis_ = axis
    # protect the new histogram dimension from being summed
    if axis is None:
        axis_ = tuple(_range(len(a.shape)))
    elif isinstance(axis, tuple) or isinstance(axis, list):
        axis_ = tuple([i - 1 if i < 0 else i for i in axis if i <= len(a.shape)])
    elif np.ndim(axis) == 0:
        if axis >= len(a.shape):
            np.sum(a, axis=axis)
            raise ValueError
        if axis == -1:
            axis_ = -2
    else:
        raise ValueError("Invalid axis value")
    return (
        np.sum(
            np.logical_or(
                np.logical_and(
                    a[..., None] >= bin_edges[:-1], a[..., None] < bin_edges[1:]
                ),
                (a[...] == bin_edges[-1])[..., None],
            ).astype(int),
            axis=axis_,
        ),
        bin_edges,
    )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import numpy as np
import rospy
from sensor_msgs.msg import OccupancyGrid

def numpy_to_occupancy_grid_msg(arr, ref_map2d, frame_id, time=None):
    if not len(arr.shape) == 2:
            raise TypeError('Array must be 2D')
    arr = arr.T * 100.
    if not arr.dtype == np.int8:
        arr = arr.astype(np.int8)
    if time is None:
        time = rospy.Time.now()
    grid = OccupancyGrid()
    grid.header.frame_id = frame_id
    grid.header.stamp.secs = time.secs
    grid.header.stamp.nsecs = time.nsecs
    grid.data = arr.ravel()
    grid.info.resolution = ref_map2d.resolution()
    grid.info.height = arr.shape[0]
    grid.info.width = arr.shape[1]
    grid.info.origin.position.x = ref_map2d.origin[0]
    grid.info.origin.position.y = ref_map2d.origin[1]
    return grid 



import pickle
import os

def freeze_test_case(func, args, kwargs={}, folder="./", case_name="testcase", omit_input=False):
    """ Automated storage of test cases

    Takes an input, and function
    runs the input through the function to get an output
    then saves both input and output to a test_data.npy file

    Func should take ndarray as input, and as output
    """
    # generate reference test output
    outputs = func(*args, **kwargs)

    # omit inputs if desired (for example if they should be loaded separately and are memory-heavy)
    if omit_input:
        args = 'omitted'
        kwargs = 'omitted'
    # save inputs and outputs to file
    data = {'args':args, 'kwargs':kwargs, 'outputs':outputs, 'funcname':func.__name__}

    folder = os.path.expandvars(os.path.expanduser(folder))
    filepath = os.path.join(folder, case_name + '.pickle')
    with open(filepath, 'wb') as f:
        pickle.dump(data, f)
    print("Tests result saved to {}".format(filepath))
    return outputs, filepath

def load_test_case(folder="./", case_name="testcase"):
    folder = os.path.expandvars(os.path.expanduser(folder))
    filepath = os.path.join(folder, case_name + '.pickle')
    with open(filepath, 'rb') as f:
        data = pickle.load(f)
    return data




import warnings
import numpy as np
from matplotlib import pyplot as plt

""" Pyplot tools
Daniel Dugas
"""
def gridshow(*args, **kwargs):
    """ Wrapper for pyplot.imshow, for use with non-image 2D arrays.

    As opposed to imshow, in gridshow:
    first dim of the array -> x axis of the plot
    second dim of the array -> y axis of the plot
    origin='lower' (doesn't flip y axis)

    All other kwargs and args are passed on as-is
    """
    if not 'origin' in kwargs:
        kwargs['origin'] = 'lower'
    return plt.imshow(*(arg.T if i == 0 else arg for i, arg in enumerate(args)), **kwargs)




# github.com/danieldugas/pyniel
import numpy as np

def pointcloud_to_voxelgrid(pointcloud, gridsize=24, n_spatial_dims=None, return_indices=False, pad=False):
    """ Transforms a pointcloud into a voxelgrid

    Parameters
    ----------
    pointcloud : array_like
        The point cloud, of shape (n_points, n_spatial_dims+n_extra_channels).
        The first n_spatial_dims values of the second dimension should
        correspond to the spatial channels x, y, z, ...
    gridsize : int, or tuple
        The shape of the grid, in i j k ... dimensions
        (corresponding to x y z ... respectively).
        if a single int, cubic grid with size_i = size_j = size_k = ...
        if a tuple, len(tuple) should be n_spatial_dims
    n_spatial_dims : int or None, optional
        if specified, will discard any dimensions past n_spatial_dims as extra_channels
    return_indices : bool, optional
        if True, returns an array of shape (n_points, n_spatial_dims), containing
        the indices in voxelgrid of each point in the original pointcloud
    pad : bool, optional
        default False,
        if True, the final result will be padded by one at the start and end of each spatial dimension
        resulting in a final grid of shape gridsize + 2

    Returns
    -------
    voxelgrid : array_like
        voxelgrid of shape (gridsize,n_spatial_dims+1)
        a channel for the occupancy of each voxel is added after after the spatial channels.
        extra_channels are discarded
    ijk : array_like
        array of shape (n_points, n_spatial_dims), containing the indices in voxelgrid of
        each point in the original pointcloud


    Example
    -------
    >>> a = np.array([[ 594.91059495,  -12.68847179,  380.85336289], [ -29.15473572,  617.57815194,  192.9789925 ], [ 617.35293371,  331.56653079,   11.62966491], [-182.31756149,  323.28363397,  -62.67164303], [ -45.40210734,  374.61999772,  236.48058249], [-238.75399034, -192.61805858,  -31.56503694], [  57.93188967,  645.13968563,    9.51061704], [ 342.32439168, -194.1729977 ,  258.29641547], [ 609.48471435,  663.82602109, -120.32760802], [ 109.49060671, -236.72820364,  269.298423  ]])
    >>> pointcloud_to_voxelgrid(a, gridsize=3)[...,-1]
    array([[[1., 0., 0.],
            [1., 0., 0.],
            [0., 1., 1.]],
    <BLANKLINE>
           [[0., 0., 1.],
            [0., 0., 0.],
            [1., 0., 0.]],
    <BLANKLINE>
           [[0., 0., 2.],
            [1., 0., 0.],
            [1., 0., 0.]]])
    """
    # Sanitize params
    if n_spatial_dims is None:
        n_spatial_dims = pointcloud.shape[1]
    if not isinstance(n_spatial_dims, int):
        raise ValueError("Expected n_spatial_dims to be an int, got {}".format(type(n_spatial_dims)))
    if isinstance(gridsize, int):
        gridsize = np.array([gridsize] * n_spatial_dims)
    else:
        gridsize = np.array(gridsize)
        if gridsize.shape != (n_spatial_dims,):
            raise ValueError(
            "Parameter gridsize must either be an int, or a tuple of length n_spatial_dims."
            "n_spatial_dims={}, gridsize.shape={}".format(n_spatial_dims, gridsize.shape))
    n_extra_channels = pointcloud.shape[1] - n_spatial_dims
    if n_extra_channels < 0:
        raise ValueError(
                "Expected pointcloud to have shape (n_points, n_spatial_dims + n_extra_channels),"
                " got {}".format(pointcloud.shape))
    # Find the xyz limits and step size
    xyzmin = pointcloud[:,:n_spatial_dims].min(axis=0)
    xyzmax = pointcloud[:,:n_spatial_dims].max(axis=0)
    dxyz = (xyzmax - xyzmin) / gridsize
    # Transform x y z ... to integer indices, normalizing axes
    ijk = ((pointcloud[:,:n_spatial_dims] - xyzmin) / dxyz).astype(int)
    # points at the border of the last voxels should be put in the last voxels
    out_of_range = (ijk == gridsize)
    ijk[out_of_range] -= 1
    # Initialize voxel grid
    # Create xyz values to fill voxel grid
    list_of_axis_ranges = [np.arange(dimsize) * dxi + xi0
            for dimsize, dxi, xi0 in zip(gridsize, dxyz, xyzmin)]
    if pad:
        list_of_axis_ranges = [np.pad(dim, (1,1), mode='reflect', reflect_type='odd')
                               for dim in list_of_axis_ranges]
    xyzgrid = np.stack(np.meshgrid(*list_of_axis_ranges, indexing="ij"), axis=-1)
    # Calculate occupancies
    ijk_unique, occupancies = np.unique(ijk, return_counts=True, axis=0)
    occupancygrid = np.zeros(tuple(gridsize))
    occupancygrid[tuple(ijk_unique.T)] = occupancies
    padsize = 0
    if pad:
        occupancygrid = np.pad(occupancygrid, 1, mode='constant', constant_values=0)
        padsize = 2
    # Concatenate xyz (n1,n2,n3,...,n_spatial_dims) and
    # occupancies (n1,n2,n3,...,1) to make full size voxelgrid
    voxelgrid = np.zeros(tuple(gridsize + padsize)+(n_spatial_dims+1,))
    voxelgrid[...,:n_spatial_dims] = xyzgrid
    voxelgrid[...,n_spatial_dims] = occupancygrid
    if return_indices:
        return voxelgrid, ijk
    else:
        return voxelgrid

def pointcloud_to_xyzgrid(pointcloud, gridsize=24, n_spatial_dims=None, pad=False):
    """ Transforms a pointcloud into an xyz grid

    Parameters
    ----------
    pointcloud : array_like
        The point cloud, of shape (n_points, n_spatial_dims+n_extra_channels).
        The first n_spatial_dims values of the second dimension should
        correspond to the spatial channels x, y, z, ...
    gridsize : int, or tuple
        The shape of the grid, in i j k ... dimensions
        (corresponding to x y z ... respectively).
        if a single int, cubic grid with size_i = size_j = size_k = ...
        if a tuple, len(tuple) should be n_spatial_dims
    n_spatial_dims : int or None, optional
        if specified, will discard any dimensions past n_spatial_dims as extra_channels
    pad : bool, optional
        default False,
        if True, the final result will be padded by one at the start and end of each spatial dimension
        resulting in a final grid of shape gridsize + 2

    Returns
    -------
    xyzgrid : array_like
        voxelgrid of shape (gridsize,n_spatial_dims)
        extra_channels are discarded


    """
    # Sanitize params
    if n_spatial_dims is None:
        n_spatial_dims = pointcloud.shape[1]
    if not isinstance(n_spatial_dims, int):
        raise ValueError("Expected n_spatial_dims to be an int, got {}".format(type(n_spatial_dims)))
    if isinstance(gridsize, int):
        gridsize = np.array([gridsize] * n_spatial_dims)
    else:
        gridsize = np.array(gridsize)
        if gridsize.shape != (n_spatial_dims,):
            raise ValueError(
            "Parameter gridsize must either be an int, or a tuple of length n_spatial_dims."
            "n_spatial_dims={}, gridsize.shape={}".format(n_spatial_dims, gridsize.shape))
    n_extra_channels = pointcloud.shape[1] - n_spatial_dims
    if n_extra_channels < 0:
        raise ValueError(
                "Expected pointcloud to have shape (n_points, n_spatial_dims + n_extra_channels),"
                " got {}".format(pointcloud.shape))
    # Transform x y z ... to integer indices, normalizing axes
    xyzmin = pointcloud[:,:n_spatial_dims].min(axis=0)
    xyzmax = pointcloud[:,:n_spatial_dims].max(axis=0)
    dxyz = (xyzmax - xyzmin) / gridsize
    # Initialize voxel grid
    # Create xyz values to fill voxel grid
    list_of_axis_ranges = [np.arange(dimsize) * dxi + xi0
            for dimsize, dxi, xi0 in zip(gridsize, dxyz, xyzmin)]
    if pad:
        list_of_axis_ranges = [np.pad(dim, (1,1), mode='reflect', reflect_type='odd')
                               for dim in list_of_axis_ranges]
    xyzgrid = np.stack(np.meshgrid(*list_of_axis_ranges, indexing="ij"), axis=-1)
    return xyzgrid


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import numpy as np
from indexing import as_idx, filter_invalid_idx


def resize_with_zeros(a, shape):
    """ special case of resize_with_fill where fill_value is 0
    """
    return resize_with_fill(a, shape, fill_value=0)


def resize_with_fill(a, shape, fill_value=np.nan):
    """ returns a, resized to shape, filling blanks with zeros

    Parameters
    ----------
    a : array_like
        Array to be reshaped.
    shape : tuple of ints
        new shape
    fill_value : value, optional
        value to fill in newly created cells

    Returns
    -------
    result : ndarray
        Array of shape (shape)

    Notes
    -----
    the new shape must have the same amount of dimensions as the old,
    as the function can not know which form of reshaping makes sense otherwise.

    Example
    -------
    >>> a = np.ones((2,3))
    >>> a
    array([[1., 1., 1.],
           [1., 1., 1.]])
    >>> resize_with_fill(a, (3,4))
    array([[ 1.,  1.,  1., nan],
           [ 1.,  1.,  1., nan],
           [nan, nan, nan, nan]])
    """
    if len(shape) != len(a.shape):
        raise ValueError(
            "Expected new shape to have as many dimensions as "
            "old shape. Reshape to desired number of dimensions "
            "before calling resize_with_fill"
        )
    resized = np.full(shape, fill_value=fill_value)
    idx = filter_invalid_idx(as_idx(a), resized)
    resized[idx] = a[idx]
    return resized


def from_list_of_lists(l, fill_value=np.nan):
    """ fill a list of lists with fill_value until all sublists are same size,
    and turn the result into an np.array

    Parameters
    ----------
    l : list
        list of lists
    fill_value : value, optional
        value to fill in newly created cells

    Returns
    -------
    result : ndarray
        Array of shape (len(l), longest)

    Example
    -------
    >>> l = [[0,1,2],[0,1],[0]]
    >>> from_list_of_lists(l)
    array([[ 0.,  1.,  2.],
           [ 0.,  1., nan],
           [ 0., nan, nan]])
    >>> l = [[[0],[1],[2]],[[0],[1]],[[0]]]
    >>> from_list_of_lists(l, fill_value=[np.nan])
    array([[[ 0.],
            [ 1.],
            [ 2.]],
    <BLANKLINE>
           [[ 0.],
            [ 1.],
            [nan]],
    <BLANKLINE>
           [[ 0.],
            [nan],
            [nan]]])
    """
    if isinstance(l[0][0], list) and not isinstance(fill_value, list):
        errorstring = (
            "fill_value must be the same type as items in the "
            "sublist. Here items in the sublist are lists of size "
            "{}, for example: {}. fill_value must also be a list "
            "of size {}."
        ).format(str(len(l[0][0])), str(l[0][0]), str(len(l[0][0])))
        raise ValueError(errorstring)
    length = len(sorted(l, key=len, reverse=True)[0])
    return np.array([list(li) + [fill_value] * (length - len(li)) for li in l])


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#
#  tsne.py
#
# Implementation of t-SNE in Python. The implementation was tested on Python 2.7.10, and it requires a working
# installation of NumPy. The implementation comes with an example on the MNIST dataset. In order to plot the
# results of this example, a working installation of matplotlib is required.
#
# The example can be run by executing: `ipython tsne.py`
#
#
#  Created by Laurens van der Maaten on 20-12-08.
#  Copyright (c) 2008 Tilburg University. All rights reserved.

import numpy as Math
import pylab as Plot


def Hbeta(D=Math.array([]), beta=1.0):
    """Compute the perplexity and the P-row for a specific value of the precision of a Gaussian distribution."""

    # Compute P-row and corresponding perplexity
    P = Math.exp(-D.copy() * beta)
    sumP = sum(P)
    H = Math.log(sumP) + beta * Math.sum(D * P) / sumP
    P = P / sumP
    return H, P


def x2p(X=Math.array([]), tol=1e-5, perplexity=30.0):
    """Performs a binary search to get P-values in such a way that each conditional Gaussian has the same perplexity."""

    # Initialize some variables
    print("Computing pairwise distances...")
    (n, d) = X.shape
    sum_X = Math.sum(Math.square(X), 1)
    D = Math.add(Math.add(-2 * Math.dot(X, X.T), sum_X).T, sum_X)
    P = Math.zeros((n, n))
    beta = Math.ones((n, 1))
    logU = Math.log(perplexity)

    # Loop over all datapoints
    for i in range(n):

        # Print progress
        if i % 500 == 0:
            print("Computing P-values for point ", i, " of ", n, "...")

        # Compute the Gaussian kernel and entropy for the current precision
        betamin = -Math.inf
        betamax = Math.inf
        Di = D[i, Math.concatenate((Math.r_[0:i], Math.r_[i + 1 : n]))]
        (H, thisP) = Hbeta(Di, beta[i])

        # Evaluate whether the perplexity is within tolerance
        Hdiff = H - logU
        tries = 0
        while Math.abs(Hdiff) > tol and tries < 50:

            # If not, increase or decrease precision
            if Hdiff > 0:
                betamin = beta[i].copy()
                if betamax == Math.inf or betamax == -Math.inf:
                    beta[i] = beta[i] * 2
                else:
                    beta[i] = (beta[i] + betamax) / 2
            else:
                betamax = beta[i].copy()
                if betamin == Math.inf or betamin == -Math.inf:
                    beta[i] = beta[i] / 2
                else:
                    beta[i] = (beta[i] + betamin) / 2

            # Recompute the values
            (H, thisP) = Hbeta(Di, beta[i])
            Hdiff = H - logU
            tries = tries + 1

        # Set the final row of P
        P[i, Math.concatenate((Math.r_[0:i], Math.r_[i + 1 : n]))] = thisP

    # Return final P-matrix
    print("Mean value of sigma: ", Math.mean(Math.sqrt(1 / beta)))
    return P


def pca(X=Math.array([]), no_dims=50):
    """Runs PCA on the NxD array X in order to reduce its dimensionality to no_dims dimensions."""

    print("Preprocessing the data using PCA...")
    (n, d) = X.shape
    X = X - Math.tile(Math.mean(X, 0), (n, 1))
    (l, M) = Math.linalg.eig(Math.dot(X.T, X))
    Y = Math.dot(X, M[:, 0:no_dims])
    return Y


def tsne(
    X=Math.array([]),
    no_dims=2,
    initial_dims=50,
    perplexity=50.0,
    max_iter=1000,
    err_threshold=0.0,
):
    """Runs t-SNE on the dataset in the NxD array X to reduce its dimensionality to no_dims dimensions.
  The syntaxis of the function is Y = tsne.tsne(X, no_dims, perplexity), where X is an NxD NumPy array."""

    # Check inputs
    if isinstance(no_dims, float):
        print("Error: array X should have type float.")
        return -1
    if round(no_dims) != no_dims:
        print("Error: number of dimensions should be an integer.")
        return -1

    # Initialize variables
    X = pca(X, initial_dims).real
    (n, d) = X.shape
    initial_momentum = 0.5
    final_momentum = 0.8
    eta = 500
    min_gain = 0.01
    Y = Math.random.randn(n, no_dims)
    dY = Math.zeros((n, no_dims))
    iY = Math.zeros((n, no_dims))
    gains = Math.ones((n, no_dims))

    # Compute P-values
    P = x2p(X, 1e-5, perplexity)
    P = P + Math.transpose(P)
    P = P / Math.sum(P)
    P = P * 4  # early exaggeration
    P = Math.maximum(P, 1e-12)

    # Run iterations
    try:
        for iter in range(max_iter):

            # Compute pairwise affinities
            sum_Y = Math.sum(Math.square(Y), 1)
            num = 1 / (1 + Math.add(Math.add(-2 * Math.dot(Y, Y.T), sum_Y).T, sum_Y))
            num[range(n), range(n)] = 0
            Q = num / Math.sum(num)
            Q = Math.maximum(Q, 1e-12)

            # Compute gradient
            PQ = P - Q
            for i in range(n):
                dY[i, :] = Math.sum(
                    Math.tile(PQ[:, i] * num[:, i], (no_dims, 1)).T * (Y[i, :] - Y), 0
                )

            # Perform the update
            if iter < 20:
                momentum = initial_momentum
            else:
                momentum = final_momentum
            gains = (gains + 0.2) * ((dY > 0) != (iY > 0)) + (gains * 0.8) * (
                (dY > 0) == (iY > 0)
            )
            gains[gains < min_gain] = min_gain
            iY = momentum * iY - eta * (gains * dY)
            Y = Y + iY
            Y = Y - Math.tile(Math.mean(Y, 0), (n, 1))

            # Compute current value of cost function
            if (iter + 1) % 10 == 0:
                C = Math.sum(P * Math.log(P / Q))
                print("Iteration ", (iter + 1), ": error is ", C)
                if C < err_threshold:
                    break

            # Stop lying about P-values
            if iter == 100:
                P = P / 4

        # Return solution
        return Y
    except (KeyboardInterrupt):
        print("Interrupted by user.")
        return Y


if __name__ == "__main__":
    print("Run Y = tsne.tsne(X, no_dims, perplexity) to perform t-SNE on your dataset.")
    print("Running example on 2,500 MNIST digits...")
    X = Math.loadtxt("mnist2500_X.txt")
    labels = Math.loadtxt("mnist2500_labels.txt")
    Y = tsne(X, 2, 50, 20.0)
    Plot.scatter(Y[:, 0], Y[:, 1], 20, labels)
    Plot.show()



import warnings
import numpy as np

""" Logic tools
Daniel Dugas
"""


def equals(tuple_of_arrays):
    """ Apply == operator to multiple arrays.
    Cells in output are true where array1 == array2 == array3 == ...

    Parameters
    ----------
    tuple_of_arrays : tuple
        contains ndarrays on which the function performs equality checks.

    Returns
    -------
    result : ndarray
        Array of type bool, with the same shape as the largest array in tuple,
        if broadcasting is successful.

    Example
    -------
    >>> A = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]])
    >>> B = np.array([[9, 1, 9], [9, 9, 1], [0, 9, 0]])
    >>> C = np.array([[6, 1, 6], [6, 9, 1], [6, 0, 0]])
    >>> equals((A, B, C))
    array([[False,  True, False],
           [False, False,  True],
           [False, False,  True]], dtype=bool)
    >>> D = np.array([0, 0 ,0])
    >>> equals((A, B, C, D))
    array([[False, False, False],
           [False, False, False],
           [False, False,  True]], dtype=bool)
    """
    truth_value = np.ones(tuple_of_arrays[0].shape)
    if len(tuple_of_arrays) <= 1:
        warnings.warn("Expected more than 1 array for logical comparison.")
        return truth_value
    for array1, array2 in zip(tuple_of_arrays[:-1], tuple_of_arrays[1:]):
        truth_value = np.logical_and(truth_value, array1 == array2)
    return truth_value


def generic_is(a, axis, generic, generic_checker=None):
    """ generic function for ismin, ismax

    Produces same result as generic function, but as truth array instead of
    indices along axis. Compatible with functions which identify a single value
    along a given axis (max, min, ...).

    e.g. ismin <- argmin(a, axis), returns an array of same shape as a
    where only min values along axis are True

    Parameters
    ----------
    a : array_like
        Array to be reshaped.
    axis : int or None
        Axis along which to apply generic. If None, the array is flattened
        before sorting.
    generic : ufunc
        function to apply to a, which outputs an array of shape
        a.shape.pop(axis) containing indices along the axis in a where the
        generic condition is satisfied.
    generic_checker : ufunc, optional
        function to apply to a for checking consistency of results.

    Returns
    -------
    result : ndarray
        Array of type bool,

    Notes
    -----
    interestingly enough,
    a[isxxx(a,axis)] is not equal to np.xxx(a,axis).flatten().
    the values will be the same in both, but their order differs.
    (see relevant example)
    I'm not sure wether this could cause buggy behavior, so be wary.

    Example
    -------
    >>> a = np.array([[0, 0, 0], [0, 0, 9], [3, 6, 0]])
    >>> a
    array([[0, 0, 0],
           [0, 0, 9],
           [3, 6, 0]])
    >>> generic_is(a, 0, np.argmax, np.max)
    array([[False, False, False],
           [False, False,  True],
           [ True,  True, False]], dtype=bool)

    as mentioned in the notes:
    >>> a[ismax(a,axis=0)]
    array([9, 3, 6])
    >>> np.max(a,axis=0).flatten()
    array([3, 6, 9])

    In a, 9 comes before 3 and 6 in the flattened index, but in np.max(a,axis),
    the matrix is squashed vertically, and 9 ends up after 3 and 6.
    """
    result = np.zeros(a.shape)
    if axis is None:
        result[np.unravel_index(generic(a, axis=None), a.shape)] = 1
    else:
        targets = generic(a, axis=axis)
        # Transform the indices along axis into indices along all a-axes
        # Why is this so ugly, is there no function for this?
        index = list(np.unravel_index(np.arange(len(targets.flatten())), targets.shape))
        index.insert(axis, targets.flatten())
        index = tuple(index)
        if generic_checker is not None:
            assert np.all(
                np.sort(a[index]) == np.sort(generic_checker(a, axis=axis).flatten())
            )
        result[index] = 1
    if generic_checker is not None:
        assert np.all(np.sum(result, axis=axis) == 1)
    return result.astype(bool)


def ismin(a, axis=None):
    """ see generic_is()
    """
    return generic_is(a, axis=axis, generic=np.argmin)


def ismax(a, axis=None):
    """ see generic_is()
    """
    return generic_is(a, axis=axis, generic=np.argmax)


def absmaximum(a, b):
    """ Elementwise maximum values from
    two equally shaped arrays a and b

    The value picked for each element is that with
    the largest absolute value.

    If the absolute values for an element are equal but
    of different signs, the sign for a is picked.

    Example
    -------
    >>> a = np.array([[-1, 0, 1], [ 2, -2, 0], [0,  3, -3]])
    >>> b = np.array([[ 3, 0,-3], [-1,  1, 0], [0, -2,  2]])
    >>> absmaximum(a, b)
    array([[ 3,  0, -3],
           [ 2, -2,  0],
           [ 0,  3, -3]])
    """
    return a * (np.abs(a) >= np.abs(b)) + b * (np.abs(b) > np.abs(a))

def absminimum(a, b):
    """ Elementwise minimum values from
    two equally shaped arrays a and b

    The value picked for each element is that with
    the smallest absolute value.

    If the absolute values for an element are equal but
    of different signs, the sign for a is picked.

    Example
    -------
    >>> a = np.array([[-1, 0, 1], [ 2, -2, 0], [0,  3, -3]])
    >>> b = np.array([[ 3, 0,-3], [-1,  1, 0], [0, -2,  2]])
    >>> absminimum(a, b)
    array([[-1,  0,  1],
           [-1,  1,  0],
           [ 0, -2,  2]])
    """
    return a * (np.abs(a) <= np.abs(b)) + b * (np.abs(b) < np.abs(a))

if __name__ == "__main__":
    import doctest

    np.random.seed(42)
    doctest.testmod()



import matplotlib
import matplotlib.pyplot as plt
from matplotlib.widgets import Button

keeps_these_objects_away_from_the_gc = []

def plotpause(interval):
    backend = plt.rcParams['backend']
    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return

def plotshow():
    """ Equivalent of plt.show(), however does not force focus on windows 
    and should exit on ctrl-c """
    plt.show()
    interval = 0.01
    try:
        while True:
            backend = plt.rcParams['backend']
            if backend in matplotlib.rcsetup.interactive_bk:
                figManager = matplotlib._pylab_helpers.Gcf.get_active()
                if figManager is not None:
                    canvas = figManager.canvas
                    if canvas.figure.stale:
                        canvas.draw()
                    canvas.start_event_loop(interval)
    except KeyboardInterrupt:
        print("KeyboardInterrupt received, exiting wait loop.")

def button_callback(event):
    print("Closing all windows")
    plt.close('all')
def plot_closeall_button():
    """ If the returned button object is not collected somewhere,
    it will be GCed and the button unresponsive. """
    fig = plt.figure("closeall_button")
    ax = plt.axes()
    bca = Button(ax, 'Close All', color=(1., 0.8, 0.8, 1))
    bca.on_clicked(button_callback)
    keeps_these_objects_away_from_the_gc.append(bca)
    return bca




import numpy as np

from .enforcetype import asarray


class Average:
    """Keeps track of the average of added values
  """

    def __init__(self, values=None, value_names=None):
        self.values = asarray(values)
        self.names = value_names
        self.name_suffix = ": "
        self.count = 0.

    def add(self, latest_values):
        latest = asarray(latest_values)
        self.count += 1.
        if self.values is None:
            self.values = latest
            if self.names is None:
                self.names = [""] * len(self.values)
                self.name_suffix = ""
            if len(self.names) != len(self.values):
                raise ValueError
        else:
            if latest.shape != self.values.shape:
                raise ValueError
            self.values = (self.values * (self.count - 1.) + latest) / self.count

    def __str__(self):
        if self.count == 0.:
            return "Not initialized"
        return ", ".join(
            [
                name + self.name_suffix + str(val)
                for name, val in zip(self.names, self.values)
            ]
        )



import sklearn as skl
import matplotlib.pyplot as plt
import numpy as np
import treelib as tl

from sklearn.utils.validation import check_is_fitted

from numpy_tools.logic import ismin

TINY = 1e-8


def euclidean_distance(a, b, axis=-1):
    return np.linalg.norm(a - b, axis=axis)


def DKL(p, q, axis=-1):
    """ KL-divergence, between distributions p and q
    """
    return np.sum(p * np.log(TINY + p / (TINY + q)), axis=axis)


def example_debug_callback(T, DAC, clusters, X):
    plt.figure()
    plt.title(T)
    c = DAC.cluster_centers_
    p = np.argmax(DAC.predict(X), axis=-1)
    plt.scatter(X[:, 0], X[:, 1], c=p)
    plt.scatter(c[:, 0], c[:, 1], marker="x", color="red")
    plt.pause(0.1)


class DeterministicAnnealingClustering(
    skl.base.BaseEstimator, skl.base.TransformerMixin
):
    """Template class for DAC

    Attributes:
        cluster_centers_ (np.ndarray): Cluster centroids y_i (n_clusters, n_features)
        cluster_probabs_ (np.ndarray): Assignment probability vectors p(y_i | x) for each sample
                                       (n_samples, n_clusters)
        bifurcation_tree_ (treelib.Tree): Tree object that contains information about cluster evolution during
                                          annealing.

    Parameters:
        n_clusters (int): Maximum number of clusters returned by DAC.
        random_state (int): Random seed.
    """

    def __init__(
        self,
        n_clusters=8,
        t_decay_rate=0.95,
        t_min=0.001,
        random_state=42,
        metric=euclidean_distance,
        log={},
    ):
        self.n_clusters = n_clusters
        self.random_state = random_state
        self.metric = metric
        self.TMIN = t_min
        self.EPS = 1e-2  # convergence tolerance for Lagrangian minimization
        self.MAX_ITER = 1000  # Max iterations for Lagrangian minimization
        self.ALPHA = t_decay_rate
        self.DELTA_SCALE = 1e-2
        self.log = {key: [] for key in log}
        # Add more parameters, if necessary.
        self.cluster_probabs_ = None
        self.bifurcation_tree_ = tl.Tree()

    def _perturb(self, y):
        delta = np.random.normal(size=y.shape)
        delta = (
            self.DELTA_SCALE
            * delta
            / (np.linalg.norm(delta, axis=-1)[..., None] + TINY)
        )
        return y + delta

    def fit(self, X, debug_callback=None, px=None):
        """Compute DAC for input vectors X

        Preferred implementation of DAC as described in reference [1].
        Consider to use initialization and reseeding as in sklearn k-means for improved performance.

        Args:
            X (np.ndarray): Input array with shape (samples, n_features)

        Returns:
            self
        """
        np.random.seed(self.random_state)

        # Scale delta to the variance of the dataset
        self.DELTA_SCALE = self.DELTA_SCALE * np.sqrt(np.var(X, axis=0))

        d_ = self.metric

        if px is None:
            px = 1.0 / len(X)  # equiprobable x
            px_2d = px
            px_3d = px
        else:
            px_2d = px[:, None]
            px_3d = px[:, None, None]
        dd = X - np.mean(X, axis=0)
        Cx = np.sum(px_3d * np.matmul(dd[:, :, None], dd[:, None, :]), axis=0)
        l, _ = np.linalg.eig(Cx)
        T0 = T = np.sqrt(np.max(l) / 2)  # dirty and wrong but works?

        self.cluster_centers_ = np.array([np.mean(X, axis=0)])
        self.bifurcation_tree_.create_node(0, 0, data={"T": T})

        last_F = -1  # init for convergence check
        while True:
            # Split clusters
            if len(self.cluster_centers_) < self.n_clusters and T > 0:
                clusters = np.copy(
                    np.concatenate(
                        (self.cluster_centers_, self.cluster_centers_), axis=0
                    )
                )
            else:
                clusters = np.copy(self.cluster_centers_)
            clusters = self._perturb(clusters)
            n_ef = len(self.cluster_centers_)
            # Optimize at T
            for i in range(self.MAX_ITER):
                # optimize pyx (c, x)
                if T > 0:
                    pyx = np.exp(-d_(X[None, :, :], clusters[:, None, :], axis=-1) / T)
                else:
                    pyx = ismin(
                        d_(X[None, :, :], clusters[:, None, :], axis=-1), axis=0
                    ).astype(float)
                Zx = np.sum(pyx, axis=0)
                pyx = pyx / (Zx + TINY)
                py = np.sum(pyx * px, axis=1)
                pxy = pyx.T * px_2d / (py + TINY)  # (x, c)
                # optimize y
                clusters = np.sum(pxy[:, :, None] * X[:, None, :], axis=0)
                # convergence check
                F = np.sum(np.log(Zx + TINY))
                diff_F = np.abs(F - last_F)
                last_F = F
                # logging
                if "distortion" in self.log:
                    D = np.sum(
                        px
                        * np.sum(
                            pyx * d_(X[None, :, :], clusters[:, None, :], axis=-1),
                            axis=0,
                        ),
                        axis=-1,
                    )
                    self.log["distortion"].append(D)
                if "temperature" in self.log:
                    self.log["temperature"].append(T)
                if "n_effective_clusters" in self.log:
                    self.log["n_effective_clusters"].append(len(self.cluster_centers_))
                if "cluster_centers" in self.log:
                    self.log["cluster_centers"].append(np.copy(self.cluster_centers_))
                # loop exit
                if diff_F < self.EPS:
                    break

            # Confirm split for most diverged cluster
            cluster_was_added = False
            if len(self.cluster_centers_) < self.n_clusters and T > 0:
                old, new = np.split(clusters, 2)
                # New clusters should have diverged from their parent but also from other old clusters
                divergence = np.min(
                    np.linalg.norm(
                        np.abs(old[None, :, :] - new[:, None, :])
                        / (self.DELTA_SCALE + TINY),
                        axis=-1,
                    ),
                    axis=1,
                )
                # New clusters should not be empty
                empty = (np.sum(pxy, axis=0) == 0)[len(old) :]
                divergence[empty] = 0

                self.cluster_centers_ = old
                if np.nanmax(divergence) > 10:
                    cluster_was_added = True
                    idx = np.argmax(divergence)
                    tag = len(self.cluster_centers_)
                    self.cluster_centers_ = np.append(
                        self.cluster_centers_, [new[idx]], axis=0
                    )
                    self.bifurcation_tree_.create_node(
                        tag, tag, parent=idx, data={"T": T}
                    )

            else:
                not_empty = np.sum(pxy, axis=0) != 0
                self.cluster_centers_[not_empty] = clusters[not_empty]
            # Optional debugging
            if debug_callback is not None:
                debug_callback(T, self, clusters, X)
            # Cooldown
            if not cluster_was_added:
                T = self.ALPHA * T
            # Exit loop
            if T == 0:
                break
            #  Check T > Tmin
            if T <= self.TMIN and len(self.cluster_centers_) == self.n_clusters:
                T = 0
                continue
        return self

    def predict(self, X):
        """Predict assignment probability vectors for each sample in X.

        Args:
            X (np.ndarray): Input array with shape (new_samples, n_features)

        Returns:
            P (np.ndarray): Assignment probability vectors (new_samples, n_clusters)
        """
        check_is_fitted(self, ["cluster_centers_"])
        d_ = self.metric
        P = ismin(
            d_(X[:, None, :], self.cluster_centers_[None, :, :], axis=-1), axis=1
        ).astype(float)
        return P

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster centers.

        Args:
            X (np.ndarray): Input array with shape (new_samples, n_features)

        Returns:
            Y (np.ndarray): Cluster-distance vectors (new_samples, n_clusters)
        """
        check_is_fitted(self, ["cluster_centers_"])

        # Your code goes here
        d_ = self.metric
        Y = d_(self.cluster_centers_[None, :, :], X[:, None, :], axis=-1)
        return Y

    def plot_bifurcation(self):
        """Show the evolution of cluster splitting"""
        check_is_fitted(self, ["bifurcation_tree_"])

        # Your code goes here

        return None

    def plot_fit(self):
        check_is_fitted(self, ["cluster_centers_"])
        try:
            dist = np.array(self.log["distortion"][:-2])
            temp = np.array(self.log["temperature"][:-2])
            n_c = np.array(self.log["n_effective_clusters"][:-2])
        except KeyError:
            raise KeyError("Logging must be enabled for plotting")
        fig = plt.figure(figsize=(20, 10))
        ax = fig.subplots(3, 1)
        ax[0].set_title("Training log")
        ax[0].plot(dist / (np.min(dist) + TINY), "k--")
        ax[0].set_ylabel("<D/Dmin>")
        ax[0].set_xscale("log", nonposx="clip")
        # ax[0].set_yscale("log", nonposy='clip')
        ax[1].step(n_c, "k")
        ax[1].set_xscale("log", nonposx="clip")
        ax[1].set_ylabel("n_effective_clusters")
        ax[2].plot(temp, "k--")
        ax[2].set_xscale("log", nonposx="clip")
        # ax[2].set_yscale("log", nonposy='clip')
        ax[2].set_ylabel("T")
        ax[2].set_xlabel("steps")
        return fig, ax


if __name__ == "__main__":
    # Generate example data
    centers = np.array([[-1, -1], [-1, 1], [1, -1], [1, 1], [0, 0]])
    X = np.concatenate(
        [np.random.normal(c, 0.5, size=(200, 2)) for c in centers], axis=0
    )
    # Fit model
    DAC = DeterministicAnnealingClustering(
        n_clusters=5,
        log={"distortion", "temperature", "n_effective_clusters", "cluster_centers"},
    )
    try:
        DAC.fit(X)
    except KeyboardInterrupt:
        print("interrupted")

    # Plot results
    c = DAC.cluster_centers_
    p = np.argmax(DAC.predict(X), axis=-1)
    fig = plt.figure(figsize=(20, 10))
    ax = fig.subplots(1, 2)
    ax[0].set_title("Input Data")
    ax[0].scatter(X[:, 0], X[:, 1])
    ax[1].set_title("Hard DAC clustering")
    ax[1].scatter(X[:, 0], X[:, 1], c=p)
    ax[1].scatter(c[:, 0], c[:, 1], marker="x", color="red")
    DAC.plot_fit()
    plt.show()



from __future__ import print_function

from matplotlib import pyplot as plt
import time
from math import floor
import numpy as np


def progress_bar(batchmaker, ax=plt.figure("progress_bar").add_subplot(111)):
    ax.scatter(time.time(), batchmaker.n_batches_consumed())
    ax.set_ylim([0, batchmaker.n_batches_consumed() + batchmaker.n_batches_remaining()])


class TextProgressBar:
    def __init__(self, width=50, silent_init=False):
        self.width = width
        self.last_consumed = 0
        if not silent_init:
            print("|" + " " * self.width + "|")
        self.max_ = None
        self.remaining_points = None

    def update(self, batchmaker):
        # max progress
        consumed = batchmaker.n_batches_consumed()
        max_ = consumed + batchmaker.n_batches_remaining()
        if self.remaining_points is not None and self.max_ != max_:
            raise Warning(
                "Maximum limit for progress has changed.\
                           Use a separate progress bar for each batchmaker."
            )
        self.max_ = max_

        # start of progress bar
        if self.remaining_points is None:
            print("|", end="")
            # a trigger value for each remaining point
            self.remaining_points = list(
                np.round(1.0 * (np.arange(self.width) + 1) * self.max_ / self.width)
            )

        # if progress has reverted, start a new bar
        if consumed < self.last_consumed:
            self.abort()

        # Find how many bar points to print
        points = len([p for p in self.remaining_points if p <= consumed])
        self.remaining_points = [p for p in self.remaining_points if p > consumed]
        print("." * points, end="")

        self.last_consumed = consumed

        # Check if bar is complete
        if len(self.remaining_points) == 0:
            self.close_and_reset()

    def abort(self):
        missing = int(floor((1.0 - self.last_consumed) * self.width)) - 1
        if missing < 0:
            raise ValueError(
                "Negative missing points should not occur,\
            possible error within batchmaker. "
                + str(self.last_consumed)
            )
        print("x" + " " * missing, end="")
        self.close_and_reset()

    def close_and_reset(self):
        print("|", end="")
        self.__init__(width=self.width, silent_init=True)



import numpy as np
from numpy_tools.indexing import as_idx


def one_hot(labels, depth, on_value=1.0, off_value=0.0):
    """ Takes an numpy array and replaces the last axis with a one_hot encoding
    Daniel Dugas

    Args:
            labels (np.ndarray): Input array
            depth (int): amount of classes in one_hot encoding

    Returns:
            (np.ndarray) one_hot encoding with shape labels.shape + (depth,)

    The inverse of one_hot is np.argmax(one_hot_array, axis=-1)

    Example
    -------
    >>> a = np.array([[0, 1, 2],[1, 2, 0], [2, 2, 1]])
    >>> a
    array([[0, 1, 2],
           [1, 2, 0],
           [2, 2, 1]])
    >>> one_hot(a, 3)
    array([[[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]],
    <BLANKLINE>
           [[0., 1., 0.],
            [0., 0., 1.],
            [1., 0., 0.]],
    <BLANKLINE>
           [[0., 0., 1.],
            [0., 0., 1.],
            [0., 1., 0.]]])
    """
    a = np.array(labels)
    one_hot = np.ones(a.shape + (depth,)) * off_value
    switch_on = as_idx(a) + (np.array(a).astype(np.intp).flatten(),)
    one_hot[switch_on] = on_value
    return one_hot


def softmax(y, axis=-1):
    """ softmax function in numpy
    applies to the last axis.
    """
    # we don't want to get rid of the sum axis, so we store the desired shape
    sumshape = list(y.shape)
    sumshape[axis] = 1
    sumshape = tuple(sumshape)
    return np.exp(y) * 1.0 / np.sum(np.exp(y), axis=axis).reshape(sumshape)



from hexgrid import *

print("Flat orientation grid examples")
print("")
print("  Cube coordinates:")
print("")
print_ascii_hex(coordinate_system="cube")
print("  Doubled coordinates:")
print("")
print_ascii_hex(v_tiles=10, h_tiles=6, coordinate_system="doubled")
print("  xy coordinates:")
print("")
print_ascii_hex(coordinate_system="xy")

print("  Arbitrary text:")
print("")
def string_creator_func(i, j):
    return ["    ",
            "o o ",
            " w  "]
print_ascii_hex(string_creator_func=string_creator_func, legend_lines=string_creator_func(0,0))



from __future__ import print_function
import numpy as np
import copy
import matplotlib.pyplot as plt

from pyniel.data_structures.graph import Graph, Node


class GraphCreatorGui(object):
    def __init__(self, figure=None, click_radius=1, debug=False):
        self.debug = debug
        self.click_radius = click_radius
        self.fig = figure
        # state variables
        self.graph = Graph()
        self.selected = None


    def run(self):
        if self.fig is None:
            self.fig = plt.figure()
        self.cp, = plt.plot([], [], 'o--', color='red')
        self.sp, = plt.plot([], [], 's--')
        cid = self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        self.update_figure()
        plt.show()

    def onclick(self, event):
        is_leftclick = event.button == 1
        is_middleclick = event.button == 2
        is_rightclick = event.button == 3
        ix, iy = event.xdata, event.ydata
        if ix is None or iy is None:
            if self.debug:
                print("invalid click")
                print(event)
            return
        # Find if an existing node is under the click
        clickednode = None
        for nodeid in self.graph.nodes:
            if np.linalg.norm(self.graph.nodes[nodeid].coords - np.array([ix, iy])) < self.click_radius:
                clickednode = nodeid
                break

        newid = None

        if is_leftclick:
            if self.selected is None:
                if clickednode is not None:
                    # select clicked node
                    if self.debug:
                        print("Selecting node")
                    self.selected = clickednode
                else:
                    if self.debug:
                        print("Adding new node and selecting it")
                    # add new node
                    newid = self.graph.get_new_id()
                    self.graph.nodes[newid] = Node(edges=set(), coords=np.array([ix, iy]))
                    self.selected = newid
            else:
                if clickednode is not None:
                    if clickednode == self.selected:
                        if self.debug:
                            print("Unselecting node")
                        # unselect clicked node
                        self.selected = None
                    else:
                        if self.debug:
                            print("Toggling edge")
                        if clickednode in self.graph.nodes[self.selected].edges:
                            # remove link between self.selected and clickednode
                            self.graph.nodes[self.selected].edges.remove(clickednode)
                            self.graph.nodes[clickednode].edges.remove(self.selected)
                        else:
                            # add link between self.selected and clickednode
                            self.graph.nodes[self.selected].edges.add(clickednode)
                            self.graph.nodes[clickednode].edges.add(self.selected)
                else:
                    if self.debug:
                        print("Adding new node connected to selected")
                    # add new node connected to self.selected
                    newid = self.graph.get_new_id()
                    self.graph.nodes[newid] = Node(edges={self.selected}, coords=np.array([ix, iy]))
                    # add link between self.selected and new node
                    self.graph.nodes[self.selected].edges.add(newid)
                    # update selection
                    self.selected = newid
        elif is_rightclick:
                if clickednode is not None:
                    if self.debug:
                        print("Removing node")
                    # remove clicked node
                    for edge in self.graph.nodes[clickednode].edges:
                        self.graph.nodes[edge].edges.remove(clickednode)
                    self.graph.nodes.pop(clickednode)
                    self.selected = None
                else:
                    if self.debug:
                        print("Unselecting node")
                    self.selected = None
        elif is_middleclick:
                if clickednode is not None:
                    if self.debug:
                        print("Selecting node")
                    self.selected = clickednode

        if self.debug:
            print(event)
            print("newid:", newid)
            print("selected:", self.selected)
            print("clicked:", clickednode)
            print("nodes:", self.graph.nodes)
            print("------------------------------")

        self.update_figure()

        if False:
            self.fig.canvas.mpl_disconnect(cid)

    def update_figure(self):
        coords = self.graph.as_2d_plot()
        # visualize nodes and edges
        if coords:
            X = np.array(coords)[:,0]
            Y = np.array(coords)[:,1]
        else: 
            X = []
            Y = []
        self.cp.set_data(X,Y)
        # visualize selection
        if self.selected is not None:
            X = [self.graph.nodes[self.selected].coords[0]]
            Y = [self.graph.nodes[self.selected].coords[1]]
        else:
            X = []
            Y = []
        self.sp.set_data(X,Y)
        # draw
        self.fig.canvas.draw()

def test_graph_creator_gui():
    """ Feed random mouse inputs to test the graph building algorithm """
    fig = plt.figure()
    plt.plot([0, 0, 1, 1, 0], [0,1, 1, 0, 0], 'k-')
    gcg = GraphCreatorGui(figure=fig, click_radius=0.05)
    gcg.cp, = plt.plot([], [], 'o--', color='red')
    gcg.sp, = plt.plot([], [], 's--')
    cl, = plt.plot([], [], 'x', color='black')
    class RandomEvent(object):
        def __init__(self):
            self.button = np.random.randint(4)
            self.xdata = np.random.rand()
            self.ydata = np.random.rand()
    for i in range(1000):
        event = RandomEvent()
        cl.set_data([event.xdata], [event.ydata])
        gcg.onclick(event)
        plt.pause(0.001)
    return 1

if __name__=="__main__":
    test_graph_creator_gui()




from __future__ import print_function
import numpy as np

"""
Cube Coordinates:

    Pointy orientation


         |
m.       |       .l
     .   |   .
        .|.
     .   |   .
 .       |       .
         |
         v
         n


        l = 0      m = 0
         \        /        /\ 
          \/\  /\/        / l\ 
          / 0\/+1\        | m| 
          |+1|| 0|        \ n/
         /\-1/\-1/\        \/
        /-1\/ 0\/+1\
        |+1 | 0 |-1 |--- n = 0
        \ 0/\ 0/\ 0/
         \/  \/ 0\/
          | 0||-1|
          \+1/\+1/
           \/  \/

    Flat orientation

  +m      -n
   \     /
    \   /
     \ /
 -----------> +l
     / \ 
    /   \ 
   /     \ 


       __            __
      / 0\          / l\ 
   __/ +1 \__      /  m \ 
  /-1\ -1 /+1\     \  n /
 / +1 \__/  0 \     \__/
 \  0 / 0\ -1 /
  \__/  0 \__/
  /-1\  0 /+1\ 
 /  0 \__/ -1 \ 
 \ +1 / 0\  0 /
  \__/ -1 \__/
     \ +1 /
      \__/

Note: a wonderful primer on hexagon grids,
https://www.redblobgames.com/grids/hexagons/

"""

def neighbors(hex_orientation="flat", coordinate_system="cube"):
    """ anti-clockwise list of neighbors starting with the highest """
    cube_neighbors = np.array([[0, 1, -1], [-1, 1, 0], [-1, 0, 1], [0, -1, 1], [1, -1, 0], [1, 0, -1]])
    if hex_orientation == "flat" and coordinate_system == "cube":
        return cube_neighbors
    elif hex_orientation == "flat" and coordinate_system == "axial":
        return cube_to_axial(cube_neighbors)
    elif hex_orientation == "flat" and coordinate_system == "doubled":
        return cube_flat_to_doubled_flat(cube_neighbors)
    elif hex_orientation == "flat" and coordinate_system == "xy":
        return cube_flat_to_xy(cube_neighbors)
    elif hex_orientation == "pointy" and coordinate_system == "xy":
        return cube_pointy_to_xy(cube_neighbors)
    else:
        raise NotImplementedError

## Coordinate transforms
def xy_to_hex(xy, hex_orientation="flat", coordinate_system="cube"):
    if hex_orientation == "flat" and coordinate_system == "cube":
        return xy_to_cube_flat(xy)
    else:
        raise NotImplementedError

def xy_to_cube_pointy(xy):
    """ Transform from cube (lmn) to 2D (xy) coordinates for Flat hexagon grid.

    Parameters
    ----------
      xy : ndarray
          points in xy, of shape (n_points, 2)

    Returns
    -------
      lmn : ndarray,
          points in lmn, of shape (n_points, 3)

    Example
    -------
    >>> xy_to_cube_pointy(np.array([[ 1. ,  0.5], [-1. ,  0.5]]))
    array([[ 1, -1,  0],
           [-1,  1,  0]])
    """
    # check constraints
    sqrt3 = np.sqrt(3.)
    # pointy hex coordinate transform
    XY_to_LM_mat = np.array([[ 1, 2/3],
                             [-1, 2/3]])
    lm = np.matmul(XY_to_LM_mat, xy.T).T.astype(int)
    lmn = axial_to_cube(lm)
    return lmn

def xy_to_cube_flat(xy):
    """ Transform from cube (lmn) to 2D (xy) coordinates for Flat hexagon grid.

    Parameters
    ----------
      xy : ndarray
          points in xy, of shape (n_points, 2)

    Returns
    -------
      lmn : ndarray,
          points in lmn, of shape (n_points, 3)

    Example
    -------
    >>> xy_to_cube_flat(np.array([[ 1. ,  0.5], [-1. ,  0.5]]))
    array([[ 1,  0, -1],
           [-1,  1,  0]])
    """
    # check constraints
    sqrt3 = np.sqrt(3.)
    # flat hex coordinate transform
    XY_to_LM_mat = np.array([[ 4/3,  0],
                             [-2/3,  1]])
    lm = np.matmul(XY_to_LM_mat, xy.T).T.astype(int)
    lmn = axial_to_cube(lm)
    return lmn

def hex_to_xy(lmn, hex_orientation="flat", coordinate_system="cube"):
    if hex_orientation == "flat" and coordinate_system == "cube":
        return cube_flat_to_xy(xy)
    else:
        raise NotImplementedError

def cube_flat_to_xy(lmn):
    """ Transform from 2D (xy) to cube (lmn) coordinates for Flat hexagon grid.

    Parameters
    ----------
      lmn : ndarray,
          points in lmn, of shape (n_points, 3)
          in order to be considered valid cube coordinates, lmn must satisfy l+m+n=0

    Returns
    ------
      xy : ndarray
          points in xy, of shape (n_points, 2)

    Example
    -------
    >>> cube_flat_to_xy(np.array([[1, 0, -1], [-1, 1, 0]]))
    array([[ 0.75,  0.5 ],
           [-0.75,  0.5 ]])
    """
    check_cube_constraint(lmn)
    sin30 = 0.5
    # flat hex coordinate transform (l and m are enough to specify x and y)
    LMN_to_XY_mat = np.array([[ 0.75, 0, 0],
                              [sin30, 1, 0]])
    xy = np.matmul(LMN_to_XY_mat, lmn.T).T
    return xy

def cube_pointy_to_xy(lmn):
    """ Transform from 2D (xy) to cube (lmn) coordinates for Flat hexagon grid.

    Parameters
    ----------
      lmn : ndarray,
          points in lmn, of shape (n_points, 3)
          in order to be considered valid cube coordinates, lmn must satisfy l+m+n=0

    Returns
    ------
      xy : ndarray
          points in xy, of shape (n_points, 2)

    Example
    -------
    >>> cube_pointy_to_xy(np.array([[1, 0, -1], [-1, 1, 0]]))
    array([[ 0.5 ,  0.75],
           [-1.  ,  0.  ]])
    """
    check_cube_constraint(lmn)
    sin30 = 0.5
    # pointy hex coordinate transform (l and n are enough to specify x and y)
    LMN_to_XY_mat = np.array([[ 1, 0, sin30],
                              [ 0, 0, -0.75]])
    xy = np.matmul(LMN_to_XY_mat, lmn.T).T
    return xy

def axial_to_cube(lm):
    """ Transforms from lm axial coordinates to lmn axial coordinates,
    adding the explicit n coordinate by solving the constraint l+m+n=0 """
    lmn = np.zeros((lm.shape[0], 3))
    lmn[:,:2] = lm
    lmn[:,2] = -lm[:,0] -lm[:,1]
    return lmn.astype(int)

def cube_to_axial(lmn):
    """ Transforms from lmn cube coordinates to lm axial coordinates,
    whereby the n is simply removed """
    return lmn[:,:2]

def check_cube_constraint(lmn):
    """ check constraints on cube grid l + m + n = 0 """
    if not np.allclose(np.sum(lmn, axis=-1), 0):
        raise ValueError("cube coordinates must satisfy l + m + n = 0")

def check_doubled_constraint(ij):
    """ check constraint on doubled grid (i + j) % 2 == 0 """
    if not np.allclose(np.mod(ij[:,0] + ij[:,1], 2), 0):
        raise ValueError("doubled coordinates must satisfy (i + j) % 2 == 0")

def cube_flat_to_doubled_flat(lmn):
    check_cube_constraint(lmn)
    ij = np.zeros((lmn.shape[0], 2))
    ij[:, 0] = lmn[:, 0]
    ij[:, 1] = 2* lmn[:, 1] + lmn[:,0]
    return ij

def doubled_flat_to_cube_flat(ij):
    check_doubled_constraint(ij)
    lmn = np.zeros((ij.shape[0], 3))
    lmn[:, 0] = ij[:, 0]
    lmn[:, 1] = 0.5 * ( ij[:, 1] - ij[:, 0] )
    lmn[:, 2] = -lmn[:,0] -lmn[:,1]
    return lmn

def print_ascii_hex(h_tiles=3, v_tiles=4,
        coordinate_system="cube", hex_orientation="flat",
        string_creator_func=None, legend_lines=None):
    """

    Parameters
    ----------
    h_tiles: int, >= 3
        number of tiles to apply horizontally
    v_tiles: int, >= 4
        number of tiles to apply vertically

    Notes
    -----
    Pattern is built out of the following horizontal/vertical repeated tiles:

    i : inner area of current hexagon
    b : border between current and neighbor hexagon
    o : neighbor (of current hexagon) inner area
    d : border between 2 neighbors of current hexagon

               11             11
     012345678901  0012345678901
    0/      \____  0doooooodbbbb
    1\      /      1doooooobiiii
    2 \____/       2odddddbiiiii
    3 /    \       3odoooobiiiii

    """
    # function which determines what gets printed in the center of the hexagons
    # takes two scalars (i, j), doubled_flat coordinates of the hexagon,
    # outputs a len 3 list of len 4 strings, (one len 4 string per writable line in the hexagon)
    if string_creator_func is None:
        def coordinates_to_str(i, j):
            # cube coordinates of bottom-right hexagon in current tile
            l, m, n = doubled_flat_to_cube_flat(np.array([[i, j]]))[0]
            if coordinate_system == "cube":
                coordinates_as_str = ["{:+4d}".format(int(cc))[:4] for cc in [l, m, n]]
            elif coordinate_system == "doubled":
                coordinates_as_str = ["{:+4d}".format(int(cc))[:4] for cc in [i, j]]
                coordinates_as_str.append('    ')
            elif coordinate_system == "xy":
                x, y = cube_flat_to_xy(np.array([[l,m,n]]))[0]
                coordinates_as_str = ["{:+.1f}".format(cc)[:4] for cc in [x, y]]
                coordinates_as_str.append('    ')
            else:
                raise NotImplementedError
            return coordinates_as_str
        string_creator_func = coordinates_to_str
    # overridable legend strings
    if legend_lines is None:
        if coordinate_system == "cube":
            coordinate_names = ['l', 'm', 'n']
        elif coordinate_system == "doubled":
            coordinate_names = ['i', 'j', ' ']
        elif coordinate_system == "xy":
            coordinate_names = ['x', 'y', ' ']
        else:
            raise NotImplementedError
        legend_lines = coordinate_names
    # print the tiles
    if hex_orientation=="flat":
        TILE_WIDTH = 12
        TILE_HEIGHT = 4
        # determine amount of lines and columns to print
        n_lines = v_tiles * TILE_HEIGHT
        n_cols = h_tiles * TILE_WIDTH
        for j_tile in range(v_tiles): # cycle tiles vertically
            for tile_line in range(TILE_HEIGHT): # print tile line-by-line
                for i_tile in range(h_tiles): # cycle tiles horizontally
                    for tile_col in range(TILE_WIDTH): # print line char-by-char
                        # don't print edge chars
                        if i_tile == (h_tiles - 1):
                            if tile_col in [8, 9, 10, 11]:
                                print(' ',end='')
                                continue
                        if j_tile == 0:
                            if i_tile == 0:
                                if (tile_line, tile_col) in [(1,0), (2,1)]:
                                    print(' ',end='')
                                    continue
                            if i_tile == (h_tiles - 1):
                                if (tile_line, tile_col) in [(1,7), (2,6)]:
                                    print(' ',end='')
                                    continue
                            if tile_line == 0 and  tile_col < 8:
                                print(' ',end='')
                                continue
                            if tile_col in [1,2,3,4,5,6] and tile_line in [0,1]:
                                print(' ',end='')
                                continue
                        if j_tile == (v_tiles - 2):
                            if tile_line == 3 and tile_col in [2,3,4,5]:
                                print(' ',end='')
                                continue
                            if i_tile == 0:
                                if tile_line == 3 and tile_col in [0, 1]:
                                    print(' ',end='')
                                    continue
                            if i_tile == (h_tiles - 1):
                                if (tile_line, tile_col) in [(3,6), (3,7)]:
                                    print(' ',end='')
                                    continue
                        if j_tile == (v_tiles - 1):
                            if i_tile == 0:
                                if (tile_line, tile_col) in [(0,0)]:
                                    print(' ',end='')
                                    continue
                            if i_tile == (h_tiles - 1):
                                if (tile_line, tile_col) in [(0,7)]:
                                    print(' ',end='')
                                    continue
                            if tile_col in [1,2,3,4,5,6]:
                                print(' ',end='')
                                continue
                            if tile_line in [1, 2, 3]:
                                print(' ',end='')
                                continue

                        # doubled flat hex coordinates of bottom-right hexagon in current tile
                        i = 2 * i_tile
                        j = 2 * (v_tiles - 1 - j_tile) - 4

                        # find which are current char is in (32 characters in tile)
                        # '/' chars
                        if (tile_line, tile_col) in [(0, 0), (1, 7), (2, 6), (3, 1)]:
                            print('/', end='')
                        # '\' chars
                        elif (tile_line, tile_col) in [(0, 7), (1, 0), (2, 1), (3, 6)]:
                            print('\\', end='')
                        # '_' chars
                        elif (tile_line, tile_col) in [(0,8), (0,9), (0,10), (0,11),
                                                       (2,2), (2,3), (2,4), (2,5)]:
                            print('_', end='')
                        # writable inner hex area
                        elif tile_line in [1, 2, 3] and tile_col == 8:
                            coordinate_index = tile_line - 1
                            print(string_creator_func(i,j)[coordinate_index], end='')
                        elif tile_line in [1, 2, 3] and tile_col in [9,10,11]:
                            print('', end='')
                        # writable top-left neighbor inner hex area
                        elif tile_line in [0, 1] and tile_col == 2:
                            i_tl = i - 1
                            j_tl = j + 1
                            coordinate_index = tile_line + 1
                            print(string_creator_func(i_tl,j_tl)[coordinate_index], end='')
                        elif tile_line in [0, 1] and tile_col in [3,4,5]:
                            print('', end='')
                        # writable bottom-left neighbor inner hex area
                        elif tile_line == 3 and tile_col == 2:
                            i_bl = i - 1
                            j_bl = j - 1
                            coordinate_index = 0
                            print(string_creator_func(i_bl,j_bl)[coordinate_index], end='')
                        elif tile_line == 3 and tile_col in [3,4,5]:
                            print('', end='')
                        # inner hex area
                        elif (tile_line in [2, 3] and tile_col == 7):
                            print(' ', end='')
                        # inner hex area of neighbors
                        elif (tile_line in [2, 3] and tile_col == 0):
                            print(' ', end='')
                        elif (tile_line in [0, 1] and tile_col in [1, 6]):
                            print(' ', end='')
                        elif (tile_line in [0, 1] and tile_col == 3):
                            print(' ', end='')
                        # errors (undefined locations)
                        else:
                            print('x', end='')

                    if i_tile == (h_tiles - 1):
                        # legend
                        if j_tile == 0:
                            print([
                            "  ____   ",
                            " /{:>4}\  ".format(legend_lines[0][:4]),
                            "/ {:>4} \ ".format(legend_lines[1][:4]),
                            "\ {:>4} / ".format(legend_lines[2][:4]),
                            ][tile_line])
                        elif j_tile == 1 and tile_line == 0:
                            print(" \____/  ")
                        # linebrea
                        else:
                            print()

if __name__ == "__main__":
    import doctest
    doctest.testmod()



import numpy as np
import tensorflow as tf


def n_dimensional_weightmul(L, W, L_shape, Lout_shape, first_dim_of_l_is_batch=True):
    """ Equivalent to matmul(W,L)
      but works for L with larger shapes than 1
      L_shape and Lout_shape are excluding the batch dimension (0)"""
    if not first_dim_of_l_is_batch:
        raise NotImplementedError
    if len(L_shape) == 1 and len(Lout_shape) == 1:
        return tf.matmul(L, W)
    # L    : ?xN1xN2xN3x...
    # Lout : ?xM1xM2xM3x...
    # W    : N1xN2x...xM1xM2x...
    # Einstein notation: letter b (denotes batch dimension)
    # Lout_blmn... = L_bijk... * Wijk...lmn...
    letters = list("ijklmnopqrst")
    l_subscripts = "".join([letters.pop(0) for _ in range(len(L_shape))])
    lout_subscripts = "".join([letters.pop(0) for _ in range(len(Lout_shape))])
    einsum_string = (
        "b"
        + l_subscripts
        + ","
        + l_subscripts
        + lout_subscripts
        + "->"
        + "b"
        + lout_subscripts
    )
    return tf.einsum(einsum_string, L, W)


def gaussian_log_likelihood(sample, mean, log_sigma_squared):
    """
    gaussian log likelihood in tensorflow
    """
    TINY = 1e-8  # Do this better
    stddev = tf.sqrt(tf.exp(log_sigma_squared))
    epsilon = (sample - mean) / (stddev + TINY)
    return tf.reduce_sum(
        -0.5 * np.log(2 * np.pi) - tf.log(stddev + TINY) - 0.5 * tf.square(epsilon),
        axis=1,
    )



import rospy
from timeit import default_timer as timer
import threading

def example_rosbag_end_callback():
    print("Executing code 1s (wall) after rosbag end.")
    return

def example_rosbag_start_callback():
    print("Executing code 1s (wall) after rosbag start.")
    return

class RosbagEndDetector(object):
    def __init__(self, rosbag_end_callback=None, rosbag_start_callback=None):
        self.latest_rostime = None
        self.rosbag_started = False
        if rosbag_start_callback is None:
            rosbag_start_callback = example_rosbag_start_callback
        if rosbag_end_callback is None:
            rosbag_end_callback = example_rosbag_end_callback
        self.rosbag_start_callback = rosbag_start_callback
        self.rosbag_end_callback = rosbag_end_callback
        # start checking in separate thread
        self.check_every_s()

    def check_every_s(self):
        if self.latest_rostime is None:
            self.latest_rostime = rospy.Time.now()
        else:
            rostime = rospy.Time.now()
            if (rostime-self.latest_rostime) <= rospy.Duration(0.05):
                if self.rosbag_started:
                    rospy.logwarn("Rosbag end/pause detected!")
                    self.rosbag_end_callback()
                    return
            else:
                if not self.rosbag_started:
                    rospy.logwarn("Rosbag started, enabling end detector.")
                    self.rosbag_started = True
                    self.rosbag_start_callback()
            self.latest_rostime = rostime
        threading.Timer(1, self.check_every_s).start()




from .divisors import divisors
import numpy as np


def gridify(
    x, y, val=None, grid=None, resolution=(100, 100), value_approximation="min"
):
    # Converts a set of points in x, y to an occupancy grid
    # Optionally a value can be attached to each point, in which case cell values are approximated
    # according to the value_approximation method
    # - average: cell value is average of values for points in cell
    # - min: cell value is min of values for points in cell
    # - max: cell value is max of values for points in cell
    # - center: cell value is value for point closest to center of cell
    # - occupancy: force reverting to occupancy values
    # In case of excessive memory use (large grid size, many points, etc)
    # the function attempts to subdivide and solve the problem recursively.
    if grid is None:
        dx = (max(x) - min(x)) / resolution[0]
        hdx = dx / 2.
        dy = (max(y) - min(y)) / resolution[1]
        hdy = dy / 2.
        grid_x = np.linspace(min(x) + hdx, max(x) - hdx, resolution[0])
        grid_y = np.linspace(min(y) + hdy, max(y) - hdy, resolution[1])
        xx, yy = np.meshgrid(grid_x, grid_y)
        xminbound = xx - hdx
        xmaxbound = xx + hdx
        yminbound = yy - hdy
        ymaxbound = yy + hdy
    else:  # TODO: in this case the grid is decentered. Option?
        grid_x = grid[0]
        grid_y = grid[1]
        resolution = (
            len(grid_x),
            len(grid_y),
        )  # NOTE: in this case resolution is coerced
        hdx = np.pad(
            np.diff(grid_x), (0, 1), "symmetric"
        )  # diff with the last element duplicated
        hdy = np.pad(np.diff(grid_y), (0, 1), "symmetric")
        xx, yy = np.meshgrid(grid_x, grid_y)
        xminbound = xx
        xmaxbound = xx + hdx[None, :]
        yminbound = yy
        ymaxbound = yy + hdy[:, None]
        # in case grid is smaller than data, remove points outside.
        crop = np.logical_or.reduce(
            (
                x < np.min(xminbound),
                x > np.max(xmaxbound),
                y < np.min(yminbound),
                y > np.max(ymaxbound),
            )
        )
        x = np.delete(x, crop)
        y = np.delete(y, crop)
        if val is not None:
            val = np.delete(val, crop)
    # C maps in which cell belongs each point - C.shape = (n, resolution)
    try:
        C = np.logical_and.reduce(
            (
                x[:, None, None] > xminbound,
                y[:, None, None] > yminbound,
                x[:, None, None] < xmaxbound,
                y[:, None, None] < ymaxbound,
            )
        )
        masked = np.ma.masked_array(val[:, None, None] * np.ones(C.shape), mask=~C)
        if val is None or value_approximation == "occupancy":
            values = np.sum(C.astype(int), axis=0)
        elif value_approximation == "average":
            values = np.average(masked, axis=0)
        elif value_approximation == "min":
            values = np.min(masked, axis=0)
        elif value_approximation == "max":
            values = np.max(masked, axis=0)
        elif value_approximation == "center":
            # TODO: would return the value of point closest to grid center
            D = np.sqrt(
                np.square(xx[None, :, :] - x[:, None, None])
                + np.square(yy[None, :, :] - y[:, None, None])
            )
            masked_D = np.ma.masked_array(D, mask=~C, fill_value=None)
            idx0 = np.argmin(masked_D, axis=0).flatten()
            idx12 = np.indices(xx.shape)
            idx1 = idx12[0].flatten()
            idx2 = idx12[1].flatten()
            values = masked[idx0, idx1, idx2].reshape(xx.shape)
    except MemoryError:
        # TODO: what happens when resolution has large primes!
        subdivisions = tuple(map(min, divisors(resolution)))
        sub_resolutions = tuple(np.array(resolution) / np.array(subdivisions))
        print(
            "Insufficient memory, splitting into %s x %s subproblems of resolution %s",
            subdivisions[0],
            subdivisions[1],
            sub_resolutions,
        )
        for sub_grid_x in np.split(xminbound[0, :], subdivisions[0]):
            for sub_grid_y in np.split(yminbound[:, 0], subdivisions[1]):
                _, sub_values = gridify(
                    x,
                    y,
                    val=val,
                    grid=(sub_grid_x, sub_grid_y),
                    value_approximation=value_approximation,
                )
                try:
                    v_values = np.vstack((v_values, sub_values))
                except:
                    v_values = sub_values
            try:
                values = np.hstack((values, v_values))
            except:
                values = np.copy(v_values)
                del v_values
        assert values.shape == xx.shape

    return (xx, yy), values



import numpy as np

# TODO: pregenerate batches, etc...


def basic_extractor(input_data, pos):
    """
    A basic extractor for extracting values from the input_data
    provided as an example.
    """
    return input_data[pos]


class Batchmaker:
    """
    Generates batches based on input_data

    Generated batches have shape (examples_per_batch, example_shape)
    and are extracted from the input_data according to extractor_func.
    In the simplest case the extractor function takes the form of basic_extractor.
    """

    def __init__(
        self,
        input_data,
        examples_per_batch,
        example_shape,
        shuffle_examples=True,
        extractor_func=basic_extractor,
    ):
        self.input_data = input_data
        self.example_shape = example_shape
        self.extractor_func = extractor_func
        # examples per batch
        if examples_per_batch is "max":
            examples_per_batch = len(input_data)
        assert type(examples_per_batch) is int
        if examples_per_batch > len(input_data):
            print(
                "WARNING: more examples per batch than possible examples in all input_data"
            )
            self.examples_per_batch = len(input_data)
        else:
            self.examples_per_batch = examples_per_batch
        # initialize example indices list
        self.remaining_example_indices = list(range(len(input_data)))
        # shuffle list if required
        if shuffle_examples:
            from random import shuffle

            shuffle(self.remaining_example_indices)
        self.batches_consumed_counter = 0

    def next_batch(self):
        assert not self.is_depleted()
        # Create a single batch
        batch_input_values = np.zeros([self.examples_per_batch] + self.example_shape)
        for i_example in range(self.examples_per_batch):
            # Create training example at index 'pos' in input_data.
            pos = self.remaining_example_indices.pop(0)
            #   input.
            batch_input_values[i_example] = np.reshape(
                self.extractor_func(self.input_data, pos), self.example_shape
            )

        self.batches_consumed_counter += 1

        return batch_input_values

    def is_depleted(self):
        return len(self.remaining_example_indices) < self.examples_per_batch

    def n_batches_remaining(self):
        return len(self.remaining_example_indices) / self.examples_per_batch

    def n_batches_consumed(self):
        return self.batches_consumed_counter


class FieldBatchmaker(Batchmaker):
    """
    A Batch contains one or several fields,
    which each are an array of shape (examples_per_batch, field_shape)
    field_shape being the shape of a single example in this field

    for example, take the case where a batch needs to be created with the following fields:
    x: shape = (16)        e.g. feature vector of size 16
    y: shape = (3)         e.g. one_hot encoding with 3 labels

    each example in this batch has a corresponding x and y field,
    for tensorflow we batch together all the x fields for several examples into an array
    of shape (examples_per_batch, 16)
    and all the y fields for the same examples into an array of shape (examples_per_batch, 3)

    In this case fields_format would be {'x': {'shape': (16,), 'extractor_func': x_extractor_func},
                                         'y': {'shape': (3,),  'extractor_func': y_extractor_func}}

    An extractor function should be provided to extract each field from the provided input data
    based on an example position.

    In the simplest case an extractor function takes the form of basic_extractor(input_data, pos).
    """

    def __init__(
        self, input_data, examples_per_batch, fields_format, shuffle_examples=True
    ):
        self.input_data = input_data
        self.fields_format = fields_format
        # examples per batch
        if examples_per_batch is "max":
            examples_per_batch = len(input_data)
        assert type(examples_per_batch) is int
        if examples_per_batch > len(input_data):
            print(
                "WARNING: more examples per batch than possible examples in all input_data"
            )
            self.examples_per_batch = len(input_data)
        else:
            self.examples_per_batch = examples_per_batch
        # initialize example indices list
        self.remaining_example_indices = list(range(len(input_data)))
        # shuffle list if required
        if shuffle_examples:
            from random import shuffle

            shuffle(self.remaining_example_indices)
        self.batches_consumed_counter = 0

    def next_batch(self):
        assert not self.is_depleted()
        batch = {
            field_name: np.zeros((self.examples_per_batch,) + field["shape"])
            for field_name, field in self.fields_format.items()
        }
        # Create a single batch
        for i_example in range(self.examples_per_batch):
            # Create training example at index 'pos' in input_data.
            pos = self.remaining_example_indices.pop(0)
            #   fill values for each field
            for field_name, field in self.fields_format.items():
                batch[field_name][i_example] = field["extractor_func"](
                    self.input_data, pos
                )
                assert batch[field_name][i_example].shape == field["shape"]

        self.batches_consumed_counter += 1

        return batch



import numpy as np
import tensorflow as tf


class ModelParams:
    """
  Base class for storing parameters of tensorflow models.
  """

    def __init__(self):
        """
    initializer provided as an example.
    should be overriden with specific default parameters for child classes.
    """
        self.INPUT_SHAPE = [16]
        self.OUTPUT_SHAPE = [3]
        self.LAYERS = [{"shape": [128]}, {"shape": [128]}, {"shape": [128]}]
        self.LEARNING_RATE = 0.001
        self.DROPOUT = 0.90  # Keep-prob
        self.FLOAT_TYPE = tf.float32
        self.DISABLE_SUMMARY = False

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)



import copy
import numpy as np
import pickle

class Node(object):
    def __init__(self, edges=set(), coords=np.array([None, None])):
        self.edges = edges
        self.coords = coords

    def __str__(self):
        return "("+str(self.edges)+", "+str(self.coords)+")"
    def __repr__(self):
        return "("+str(self.edges)+", "+str(self.coords)+")"

class Graph(object):
    """ Directed graph """
    def __init__(self):
        self.nodes = {} # { nodeid: node, ... }

    def __str__(self):
        result = "Graph:\n"
        for key in self.nodes:
            result += "  ({:>5.1f}, {:>5.1f}) {} ->".format(
                self.nodes[key].coords[0], self.nodes[key].coords[1], key)
            for e in self.nodes[key].edges:
                result += " {}".format(e)
            result += "\n"
        return result
    def __repr__(self):
        return self.nodes.__repr__()

    def get_new_id(self):
        if self.nodes:
            return sorted(list(self.nodes))[-1]+1
        else:
            return 0

    def as_2d_plot(self):
        # list of coordinates from graph edges by walking through each edge
        tempnodes = copy.deepcopy(self.nodes)
        coords = []
        while tempnodes:
            coords.append([np.nan, np.nan])
            start = list(tempnodes)[0]
            coords.append(tempnodes[start].coords)
            if tempnodes[start].edges:
                end = tempnodes[start].edges.pop()
                tempnodes[end].edges.remove(start)
                coords.append(tempnodes[end].coords)
            else:
                tempnodes.pop(start)
        return coords

    def reassign_ids(self):
        # map oldnodeid to new nodeid
        map_old_to_new = {}
        for i, key in enumerate(self.nodes):
            map_old_to_new[key] = i
        # apply map to graph
        newnodes = {}
        for key in self.nodes:
            newkey = map_old_to_new[key]
            newnodes[newkey] = self.nodes[key]
            oldedges = self.nodes[key].edges
            # clear edges and replace with newkeys
            newnodes[newkey].edges = set()
            for e in oldedges:
                newnodes[newkey].edges.add(map_old_to_new[e])
        self.nodes = newnodes

    def save_to_file(self, filepath):
        with open(filepath,"wb") as f:
            pickle.dump(self.nodes, f)
        print("Graph saved to {}".format(filepath))

    def restore_from_file(self, filepath):
        with open(filepath,"rb") as f:
            self.nodes = pickle.load(f)
        print("Graph loaded from {}".format(filepath))




import numpy as np

""" Arithmetic progressions
Daniel Dugas
"""


def divisors(n):
    """ lists all the divisors for the number n

    Example
    -------
    >>> divisors(20)
    array([ 2,  4,  5, 10])
    """
    x = np.abs(n)
    candidates = np.arange(2, np.floor(np.sqrt(x)) + 1)
    divisors = candidates[np.equal(np.mod(x, candidates), 0)]
    divisors2 = x * 1. / divisors[::-1]
    if divisors[-1] == divisors2[0]:
        divisors2 = divisors2[1:]
    return np.concatenate((divisors, divisors2), axis=0).astype(int)


def primes(n):
    """ Lists primes smaller than n
    using the sieve of Erasthothenes
    (half sieve, multiples of 2 already removed)

    Example
    -------
    >>> primes(120)
    array([  2,   3,   5,   7,  11,  13,  17,  19,  23,  29,  31,  37,  41,
            43,  47,  53,  59,  61,  67,  71,  73,  79,  83,  89,  97, 101,
           103, 107, 109, 113])
    """
    sieve = np.full((n // 2,), True)
    for i in range(3, int(np.floor(np.sqrt(n))) + 1, 2):
        if sieve[(i - 1) // 2]:
            sieve[(i - 1) // 2 + i :: i] = False
    primes = np.where(sieve)[0] * 2 + 1
    primes[0] = 2
    return primes


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import print_function
import rosbag
import sys

rosbag_path = sys.argv[1]
messages = rosbag.Bag(rosbag_path, 'r').read_messages(raw=True)

topic_size_dict = {}
for topic, msg, time in messages:
    msg_size = len(msg[1])
    topic_size_dict[topic] = topic_size_dict.get(topic, 0) + msg_size
topic_size = list(topic_size_dict.items())
topic_size.sort(key=lambda x: x[1])
for topic, size in topic_size[::-1]:
    print(topic, size / 1000000., "MB")
print("----------------------------")
print("total: ", sum([size for _, size in topic_size]))



import numpy as np


def asarray(x):
    """
  converts scalars or python lists to np.array
  None outputs to None
  """
    if x is None:
        return None
    y = np.array(x)
    if len(y.shape) == 0:
        y = np.array([x])
    return y



import os

def make_dir_if_not_exists(dir_):
    try: 
        os.makedirs(dir_)
    except OSError:
        if not os.path.isdir(dir_):
            raise

def crawl_subdirs_for_files(dir_, extension=".jpg"):
    files = []
    for dirpath, dirnames, filenames in os.walk(dir_):
        for filename in [f for f in filenames if f.endswith(extension)]:
            files.append(os.path.join(dirpath, filename))
    return files



import numpy as np
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from matplotlib import pyplot as plt

# same as Greys, but shrunk and cycled 10 times over the range
Greys_c10 = ListedColormap(plt.cm.Greys(np.mod(np.linspace(0., 10., 256), 1.)))

# Red to green through yellow
trafficlight = LinearSegmentedColormap.from_list("", ["tomato","yellow","springgreen"])




import warnings
import numpy as np

from sensor_msgs.msg import PointCloud2
from sensor_msgs import point_cloud2 as pc2
""" ros pointcloud tools
Daniel Dugas
"""


def pointcloud2_to_numpy(pointcloud, field_names=None):
    """ Convert from ros PointCloud2 Message to numpy array.
    preserves structure (height, width)

    Parameters
    ----------
    pointcloud : PointCloud2
        point cloud message to convert
    field_names : list or None
        default None. Keeps all fields.
        if list of strings, fetches only fields where name is in field_names.
        ordering of fields in output array is same as order in PointCloud2 fields.

    Returns
    -------
    result : ndarray
        converted array.
        shape is (width, n_fields) if pointcloud.height is 1
        shape is (height, width, n_fields) otherwise

    Example
    -------
    """
    assert isinstance(pointcloud, PointCloud2)
    if field_names is None:
        field_names = [f.name for f in pointcloud.fields]
    result = np.array([list(point) for point in pc2.read_points(pointcloud, field_names=field_names)])
    if pointcloud.height > 1:
        result = result.reshape((pointcloud.height, pointcloud.width, -1))
    return result



if __name__ == "__main__":
    import doctest
    doctest.testmod()



import numpy as np

def chunks_split(array, chunksize, axis=0):
    """
    split array into equally sized chunks of size chunksize

    outputs:
    -----
    chunks: list of equally shaped chunks, which have similar shape to input array except for axis, which has size chunksize
    remainder: similar shape to input array except for axis, which has size in range [0,chunksize]
    """
    n_chunks = array.shape[axis] // chunksize
    splittable, remainder  = np.split(array, [n_chunks*chunksize], axis=axis)
    chunks = np.split(splittable, n_chunks, axis=axis)
    return chunks, remainder




import time

class WalltimeRate(object):
    """ Convenience class for reproducing rostime Rate with walltime """
    def __init__(self, frequency, keep_phase=False):
        if frequency <= 0:
            raise ValueError("Frequency can not be <= 0")
        self.last_time = time.time()
        self.sleep_duration = 1./frequency
        self.keep_phase = keep_phase

    def _time_to_next_and_sleep(self, sleep=False):
        current_time = time.time()
        elapsed = current_time - self.last_time
        # moved back in time
        if elapsed < 0:
            self.last_time = current_time
            time_to_next = 0
            if sleep:
                # sleep(0)
                self.last_time = current_time
            return time_to_next
        # missed a few cycles
        elif elapsed > self.sleep_duration:
            time_to_next = 0
            if sleep:
                # sleep(0)
                if self.keep_phase:
                # update last_time to the start of the current bin, violating frequency to
                # keep the original phase.
                # |     |     |     |     |     |  x  x <- resulting next scheduled timeout
                # ^start                           ^now
                #             ^last_time        ^'new' last_time
                    n_bins_to_skip = int(elapsed / self.sleep_duration)
                    self.last_time = self.last_time + n_bins_to_skip * self.sleep_duration
                else:
                    self.last_time = current_time
            return time_to_next
        # normal situation
        else:
            time_to_next = max(self.sleep_duration - elapsed, 0)
            if sleep:
                time.sleep(time_to_next)
                self.last_time += time_to_next
            return time_to_next

    def remaining(self):
        return self._time_to_next_and_sleep(sleep=False)

    def sleep(self):
        return (self._time_to_next_and_sleep(sleep=True) == 0)



import numpy as np
import tensorflow as tf

from .modelparams import ModelParams
from .custom_ops import n_dimensional_weightmul


class FCNParams(ModelParams):
    def __init__(self):
        self.INPUT_SHAPE = [16]
        self.OUTPUT_SHAPE = [3]
        self.FC_LAYERS = [{"shape": [128]}, {"shape": [128]}, {"shape": [128]}]
        self.LEARNING_RATE = 0.001
        self.DROPOUT = 0.90  # Keep-prob
        self.FLOAT_TYPE = tf.float32
        self.DISABLE_SUMMARY = False


TINY = 1e-8


class FCNetwork(object):
    def __init__(self, model_params):
        self.MP = model_params

        tf.reset_default_graph()
        preset_batch_size = None
        self.variables = []
        self.zero = tf.constant(0)
        self.global_step_tensor = tf.Variable(1, trainable=False, name="global_step")
        self.variables.append(self.global_step_tensor)
        # Graph input
        with tf.name_scope("Placeholders") as scope:
            self.input_placeholder = tf.placeholder(
                self.MP.FLOAT_TYPE,
                shape=[preset_batch_size] + self.MP.INPUT_SHAPE,
                name="input",
            )
            self.ground_truth_placeholder = tf.placeholder(
                self.MP.FLOAT_TYPE,
                shape=[preset_batch_size] + self.MP.OUTPUT_SHAPE,
                name="ground_truth",
            )
            if self.MP.DROPOUT is not None:
                default_dropout = tf.constant(1, dtype=self.MP.FLOAT_TYPE)
                self.dropout_placeholder = tf.placeholder_with_default(
                    default_dropout, (), name="dropout_prob"
                )
        # Input Layer
        previous_layer = self.input_placeholder
        previous_layer_shape = (
            self.MP.INPUT_SHAPE
        )  # Excludes batch dim (which should be at pos 0)
        # Fully connected Layers
        for i, LAYER in enumerate(self.MP.FC_LAYERS):
            previous_layer, previous_layer_shape = self.build_FC_layer(
                LAYER,
                previous_layer,
                previous_layer_shape,
                "Layer" + str(i),
                "Layer" + str(i) + "Weights",
                "_layer_" + str(i),
                activation=tf.nn.relu,
            )

        self.output, _ = self.build_FC_layer(
            {"shape": self.MP.OUTPUT_SHAPE},
            previous_layer,
            previous_layer_shape,
            "Output",
            "OutputWeights",
            "_output_layer",
            activation=tf.nn.relu,
        )

        # Loss
        with tf.name_scope("Loss") as scope:
            # Cross entropy loss of output probabilities vs. input certainties.
            self.loss = tf.losses.softmax_cross_entropy(
                self.ground_truth_placeholder, self.output
            )
            # Average sum of costs over batch.
            self.cost = tf.reduce_mean(self.loss)
            if not self.MP.DISABLE_SUMMARY:
                tf.summary.scalar("cost", self.cost)

        # Optimizers (ADAM)
        with tf.name_scope("Optimizer") as scope:
            self.optimizer = tf.train.AdamOptimizer(learning_rate=self.MP.LEARNING_RATE)
            self.train_op = self.optimizer.minimize(
                self.cost, global_step=self.global_step_tensor
            )
        # Initialize session
        self.catch_nans = tf.add_check_numerics_ops()
        self.sess = tf.Session()
        self.merged = (
            tf.summary.merge_all() if not self.MP.DISABLE_SUMMARY else self.zero
        )
        tf.global_variables_initializer().run(session=self.sess)
        # Saver
        self.saver = tf.train.Saver(self.variables)
        tf.get_default_graph().finalize()

    def variable_summaries(self, var):
        """Attach a lot of summaries to a Tensor (for TensorBoard visualization)."""
        if not self.MP.DISABLE_SUMMARY:
            with tf.name_scope("summaries"):
                mean = tf.reduce_mean(var)
                tf.summary.scalar("mean", mean)
                with tf.name_scope("stddev"):
                    stddev = tf.sqrt(tf.reduce_mean(tf.square(var - mean)))
                tf.summary.scalar("stddev", stddev)
                tf.summary.scalar("max", tf.reduce_max(var))
                tf.summary.scalar("min", tf.reduce_min(var))
                tf.summary.histogram("histogram", var)

    def build_FC_layer(
        self,
        LAYER,
        previous_layer,
        previous_layer_shape,
        scope_name,
        varscope_name,
        var_suffix="",
        reuse=False,
        activation=tf.nn.softplus,
        put_variables_in_list=None,
    ):
        with tf.name_scope(scope_name) as scope:
            layer_shape = LAYER["shape"]
            with tf.variable_scope(varscope_name, reuse=reuse) as varscope:
                weights = tf.get_variable(
                    "weights" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=previous_layer_shape + layer_shape,
                    initializer=tf.contrib.layers.xavier_initializer(),
                )
                biases = tf.get_variable(
                    "biases" + var_suffix,
                    dtype=self.MP.FLOAT_TYPE,
                    shape=layer_shape,
                    initializer=tf.constant_initializer(0),
                )
            if not reuse:
                self.variables.append(weights)
                self.variables.append(biases)
                if put_variables_in_list is not None:
                    put_variables_in_list.append(weights)
                    put_variables_in_list.append(biases)
                self.variable_summaries(weights)
            layer_output = tf.add(
                n_dimensional_weightmul(
                    previous_layer, weights, previous_layer_shape, layer_shape
                ),
                biases,
            )
            layer_output = activation(layer_output)
            if self.MP.DROPOUT is not None:
                layer_output = tf.nn.dropout(layer_output, self.dropout_placeholder)
        return layer_output, layer_shape

    ## Example functions for different ways to call the model graph.
    def classify_batch(self, batch_input):
        return self.sess.run(
            self.output, feed_dict={self.input_placeholder: batch_input}
        )

    def train_on_single_batch(
        self,
        batch_input,
        batch_ground_truth,
        train_target=None,
        cost_only=False,
        dropout=None,
        summary_writer=None,
    ):
        # feed placeholders
        dict_ = {
            self.input_placeholder: batch_input,
            self.ground_truth_placeholder: batch_ground_truth,
        }
        if self.MP.DROPOUT is not None:
            dict_[self.dropout_placeholder] = (
                self.MP.DROPOUT if dropout is None else dropout
            )
        else:
            if dropout is not None:
                raise ValueError(
                    "This model does not implement dropout yet a value was specified"
                )
        # Graph nodes to target
        cost = [self.cost]
        opt = train_target if train_target is not None else self.train_op
        # compute
        cost, _, _, summary = self.sess.run(
            (cost, opt, self.catch_nans, self.merged), feed_dict=dict_
        )
        if summary_writer is not None:
            summary_writer.add_summary(
                summary, tf.train.global_step(self.sess, self.global_step_tensor)
            )
        return np.array(cost)

    def cost_on_single_batch(
        self, batch_input, batch_ground_truth, summary_writer=None
    ):
        return self.train_on_single_batch(
            batch_input,
            batch_ground_truth,
            train_target=self.zero,
            dropout=1.0,
            summary_writer=summary_writer,
        )



import warnings
import numpy as np

""" Index tools
Daniel Dugas
"""


def as_idx_array(a, axis=None):
    """ Returns an array of shape a containing indices for a

    Parameters
    ----------
    a : array_like
        Array to be reshaped.
    axis : int, None, list of ints, or 'all'
        Axis along which to return indices. If None, the flat index.

    Returns
    -------
    result : ndarray
        Array of shape (a.shape, len(axis))
        if axis is None or a single int, (a.shape,)

    Notes
    -----

    Example
    -------
    >>> a = np.array([[0, 0, 0], [0, 0, 9], [3, 6, 0]])
    >>> a
    array([[0, 0, 0],
           [0, 0, 9],
           [3, 6, 0]])
    >>> as_idx_array(a)
    array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])
    >>> as_idx_array(a, axis='all')
    array([[[0, 0],
            [0, 1],
            [0, 2]],
    <BLANKLINE>
           [[1, 0],
            [1, 1],
            [1, 2]],
    <BLANKLINE>
           [[2, 0],
            [2, 1],
            [2, 2]]])
    >>> as_idx_array(a, axis=0)
    array([[0, 0, 0],
           [1, 1, 1],
           [2, 2, 2]])
    """
    if axis is None:
        return np.arange(len(a.flatten())).reshape(a.shape)
    idxs = np.array(np.where(np.ones(a.shape))).T.reshape(a.shape + (-1,))
    if axis == "all":
        return idxs
    return idxs[..., axis]


def as_idx(a):
    """ returns a tuple with indices for all values in a

    Parameters
    ----------
    a : array_like
        Array to be reshaped.

    Returns
    -------
    result : tuple of ndarrays
        of size len(a.shape)
        each ndarray has shape (len(a.flatten()),)

    Example
    -------
    >>> a = np.array([[0, 0, 0], [0, 0, 9], [3, 6, 0]])
    >>> a
    array([[0, 0, 0],
           [0, 0, 9],
           [3, 6, 0]])
    >>> as_idx(a)
    (array([0, 0, 0, 1, 1, 1, 2, 2, 2]), array([0, 1, 2, 0, 1, 2, 0, 1, 2]))
    >>> np.all(a[as_idx(a)] == a.flatten())
    True
    """
    return np.where(np.ones(a.shape))


def filter_invalid_idx(idx, array, ignore_extra_dimensions=False):
    """ removes indices from idx which do not point to a location in array

    Parameters
    ----------
    idx : tuple of indices
        indices to verify
    array : ndarray
        array in which indices are tested
    ignore_extra_dimensions : bool
        if True, idx values for axes not in array are ignored, and do not
        prevent indices from being valid if the other axes are within array.
        if False, all indices in idx with more dimensions than a are strictly
        marked as invalid.

    Returns
    -------
    result : tuple of indices
        subset of idx with only valid values

    Example
    -------
    >>> a = np.ones((1,2,3))
    >>> b = np.ones((3,2,1))
    >>> idx = np.where(b==1)
    >>> idx
    (array([0, 0, 1, 1, 2, 2]), array([0, 1, 0, 1, 0, 1]), array([0, 0, 0, 0, 0, 0]))
    >>> filter_invalid_idx(idx, a)
    (array([0, 0]), array([0, 1]), array([0, 0]))
    """
    if not ignore_extra_dimensions:
        if len(array.shape) < len(idx):
            warnings.warn(
                "Indices point to more dimensions than exist in "
                "array. Default behavior is to flag them all as "
                "invalid. Set ignore_extra_dimensions to True to "
                "reduce strictness. Otherwise ensure that the "
                "dimensionalities match."
            )
            return tuple()
    maxdim = min(len(array.shape), len(idx))  # larger dimensions are irrelevant
    max_idx = np.array(array.shape[:maxdim]) - 1  # max index for each axis
    test = np.array(idx[:maxdim]).T
    is_valid = np.all(np.logical_and(test <= max_idx, test >= -1 - max_idx), axis=-1)
    return tuple(test[is_valid].T)


def sliding_window(a, size=3, fill_value=np.nan):
    """ Applies a sliding window of shape size^a_ndim to array a,

    TODO: generalize to a shape > 2x2
    returns an array of shape (size^a_ndim, a.shape)
    in which the first dimension corresponds to value of a for each flat index
    in the window."""
    # We generate pad_widths for each
    paddings = np.array(np.where(np.ones((size,) * a.ndim))).T[:, :, None] * [1, -1] + [
        0,
        size - 1,
    ]
    paddeds = np.array(
        [np.pad(a, pad, "constant", constant_values=fill_value) for pad in paddings]
    )
    return paddeds[:, size // 2 : -(size // 2), size // 2 : -(size // 2)]

def filter_if_out_of_bounds(indices_list, a):
    valid_mask = np.all(
            np.logical_and(indices_list >= 0, indices_list < a.shape),
            axis=-1,
            )
    return indices_list[valid_mask]

def batchsplit(x, batchsize, axis=0):
    """ similar to np.split, but takes a batch size instead of a 
    number of splits as input
    """
    if axis != 0:
        raise NotImplementedError
    n_full_sections =  int(np.floor(len(x) / batchsize))
    if n_full_sections == 0:
        return [x]
    indices = [n * batchsize for n in range(1, n_full_sections+1)]
    sections = np.split(x, indices, axis=axis)
    if len(sections[-1]) == 0:
        sections = sections[:-1]
    return sections

if __name__ == "__main__":
    import doctest

    doctest.testmod()
